import pytest
from typing import Optional

from goodai.ltm.mem.chunk import Chunk


@pytest.fixture
def default_chunk():
    return Chunk(chunk_id=1, capacity=10, from_token_seq_id=0, metadata=None, importance=None, timestamp=123.456)


# Tests start here

def test_chunk_initialization(default_chunk):
    chunk = default_chunk
    assert chunk.chunk_id == 1
    assert chunk.capacity == 10
    assert chunk.from_token_seq_id == 0
    assert chunk.to_token_seq_id == 0
    assert chunk.indexed_length == -1
    assert chunk.timestamp == 123.456
    assert chunk.associated_keys == []


def test_chunk_length(default_chunk):
    chunk = default_chunk
    assert len(chunk) == 0

    chunk.to_token_seq_id = 5
    assert len(chunk) == 5


def test_chunk_room(default_chunk):
    chunk = default_chunk
    assert chunk.get_room() == 10

    chunk.to_token_seq_id = 5
    assert chunk.get_room() == 5


def test_chunk_at_capacity(default_chunk):
    chunk = default_chunk
    assert not chunk.is_at_capacity()

    chunk.to_token_seq_id = 10
    assert chunk.is_at_capacity()


def test_chunk_extend_by(default_chunk):
    chunk = default_chunk
    chunk.extend_by(5)
    assert chunk.to_token_seq_id == 5

    with pytest.raises(ValueError):
        chunk.extend_by(6)  # This should exceed capacity


def test_chunk_shift(default_chunk):
    chunk = default_chunk
    chunk.shift(5)
    assert chunk.from_token_seq_id == 5
    assert chunk.to_token_seq_id == 5


def test_chunk_indexed_state(default_chunk):
    chunk = default_chunk
    assert not chunk.is_indexed()

    chunk.update_indexed_state()
    assert chunk.is_indexed()


def test_chunk_add_key(default_chunk):
    chunk = default_chunk
    chunk.add_key(100)
    assert 100 in chunk.associated_keys


# Additional tests for Chunk creation and constructor

def test_chunk_creation_with_defaults():
    chunk = Chunk(chunk_id=2, capacity=20, from_token_seq_id=5, metadata=None, importance=None, timestamp=456.789)

    assert chunk.chunk_id == 2
    assert chunk.capacity == 20
    assert chunk.from_token_seq_id == 5
    assert chunk.to_token_seq_id == 5  # Should be initialized to from_token_seq_id
    assert chunk.indexed_length == -1
    assert chunk.timestamp == 456.789
    assert chunk.associated_keys == []
    assert chunk.importance is None
    assert chunk.metadata is None


def test_chunk_creation_with_metadata_and_importance():
    metadata = {"key": "value"}
    importance = 0.75
    chunk = Chunk(chunk_id=3, capacity=30, from_token_seq_id=10, metadata=metadata, importance=importance,
                  timestamp=789.012)

    assert chunk.chunk_id == 3
    assert chunk.capacity == 30
    assert chunk.from_token_seq_id == 10
    assert chunk.to_token_seq_id == 10
    assert chunk.indexed_length == -1
    assert chunk.timestamp == 789.012
    assert chunk.associated_keys == []
    assert chunk.importance == 0.75
    assert chunk.metadata == {"key": "value"}


def test_chunk_creation_with_defaults_no_keywords():
    chunk = Chunk(2, 20, 5, None, None, 456.789)

    assert chunk.chunk_id == 2
    assert chunk.capacity == 20
    assert chunk.from_token_seq_id == 5
    assert chunk.to_token_seq_id == 5  # Should be initialized to from_token_seq_id
    assert chunk.indexed_length == -1
    assert chunk.timestamp == 456.789
    assert chunk.associated_keys == []
    assert chunk.importance is None
    assert chunk.metadata is None


def test_chunk_creation_with_metadata_and_importance_no_keywords():
    metadata = {"key": "value"}
    importance = 0.75
    chunk = Chunk(3, 30, 10, metadata, importance, 789.012)

    assert chunk.chunk_id == 3
    assert chunk.capacity == 30
    assert chunk.from_token_seq_id == 10
    assert chunk.to_token_seq_id == 10
    assert chunk.indexed_length == -1
    assert chunk.timestamp == 789.012
    assert chunk.associated_keys == []
    assert chunk.importance == 0.75
    assert chunk.metadata == {"key": "value"}



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable

import msgpack

from goodai.ltm.mem.auto import AutoTextMemory
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.persistence.abstract import MemoryPersistence
from goodai.ltm.persistence.json_persistence import save_vector_db_to_npz, load_vector_db_from_npz, \
    load_chunk_queue_from_dict, load_vector_db_from_dict, default_obj_to_dict, default_dict_to_obj


def serialize_to_msgpack(obj: Any, obj_to_dict: Callable) -> bytes:
    return msgpack.packb(obj, default=obj_to_dict, use_bin_type=True)


@dataclass
class MsgPackMemoryPersistence(MemoryPersistence):
    """
    Saves and loads the whole memory as a msgpack file. Similar to json, but much more compact.
    """
    chunk_queue_filename: str = "chunk_queue.msgpack"
    vector_db_npz_filename = "vector_db.msgpack"
    obj_to_dict: Callable = default_obj_to_dict
    dict_to_obj: Callable = default_dict_to_obj

    def save(self, memory: DefaultTextMemory, directory: Path):
        chunk_queue = memory.chunk_queue
        file = directory / self.chunk_queue_filename
        with open(file, "wb") as fd:
            fd.write(serialize_to_msgpack(chunk_queue, self.obj_to_dict))
        vector_db = memory.vector_db
        file = directory / self.vector_db_npz_filename
        with open(file, "wb") as fd:
            fd.write(serialize_to_msgpack(vector_db, self.obj_to_dict))

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.chunk_queue_filename
        with open(file, "rb") as data:
            chunk_queue = load_chunk_queue_from_dict(
                msgpack.unpackb(data.read(), raw=False, strict_map_key=False, object_hook=self.dict_to_obj)
            )

        file = directory / self.vector_db_npz_filename
        with open(file, "rb") as data:
            vector_db = load_vector_db_from_dict(msgpack.unpackb(data.read(), raw=False, strict_map_key=False))

        return AutoTextMemory.create(chunk_queue=chunk_queue, vector_db=vector_db, **kwargs)

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.chunk_queue_filename,
            directory / self.vector_db_npz_filename,
        ]
        return all([file.exists() for file in files])


@dataclass
class MsgNpzPackMemoryPersistence(MemoryPersistence):
    chunk_queue_filename: str = "chunk_queue.msgpack"
    vector_db_npz_filename = "vector_db.npz"
    obj_to_dict: Callable = default_obj_to_dict
    dict_to_obj: Callable = default_dict_to_obj

    def save(self, memory: DefaultTextMemory, directory: Path):
        chunk_queue = memory.chunk_queue
        file = directory / self.chunk_queue_filename
        with open(file, "wb") as fd:
            fd.write(serialize_to_msgpack(chunk_queue, self.obj_to_dict))
        vector_db = memory.vector_db
        file = directory / self.vector_db_npz_filename
        save_vector_db_to_npz(vector_db, file)

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.chunk_queue_filename
        with open(file, "rb") as data:
            chunk_queue = load_chunk_queue_from_dict(
                msgpack.unpackb(data.read(), raw=False, strict_map_key=False, object_hook=self.dict_to_obj)
            )

        file = directory / self.vector_db_npz_filename
        vector_db = load_vector_db_from_npz(file)
        return AutoTextMemory.create(chunk_queue=chunk_queue, vector_db=vector_db, **kwargs)

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.chunk_queue_filename,
            directory / self.vector_db_npz_filename,
        ]
        return all([file.exists() for file in files])



import gc
import math
from typing import List, Tuple, Optional

import numpy as np
import torch
from goodai.ltm.data.query_passage.dataset import QueryPassageDataset

from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource

from goodai.ltm.data.query_passage.auto_data_source import AutoQueryPassageDataSource
from torch import nn
from torch.utils.data import DataLoader, Dataset
from transformers import PreTrainedTokenizer
from timeit import default_timer as timer

from goodai.ltm.reranking.prob_model import BaseQueryPassageProbModel
from goodai.helpers.sched_opt import ScheduledOptimizer


class QPPMTrainer:
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                 num_epochs: int, switch_ds_every: int, num_ds_examples: int, batch_size: int,
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 track_validation: bool, lm_lr: float, extras_lr: float, device: torch.device,
                 num_warmup_steps: int = 0, weight_decay: float = 1e-3, approx_positive_fraction: float = 0.5):
        super().__init__()
        self.approx_positive_fraction = approx_positive_fraction
        self.weight_decay = weight_decay
        self.num_warmup_steps = num_warmup_steps
        self.extras_lr = extras_lr
        self.lm_lr = lm_lr
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.track_validation = track_validation
        self.batch_size = batch_size
        self.num_ds_examples = num_ds_examples
        self.switch_ds_every = switch_ds_every
        self.num_epochs = num_epochs
        self.tokenizer = tokenizer
        self.random = random
        self.device = device
        self.train_data_sources: List[Tuple[BaseQueryPassageDataSource, float]] = []
        self.valid_data_sources: List[Tuple[BaseQueryPassageDataSource, float]] = []

    def add_data_source(self, ds_name: str, weight: float):
        train_ds, valid_ds = AutoQueryPassageDataSource.create(ds_name, self.random, self.tokenizer,
                                                               max_query_tokens=self.max_query_tokens,
                                                               min_passage_tokens=self.min_passage_tokens,
                                                               max_passage_tokens=self.max_passage_tokens)
        self.train_data_sources.append((train_ds, weight,))
        self.valid_data_sources.append((valid_ds, weight,))

    def train_dataset(self, dataset: Dataset, model: BaseQueryPassageProbModel,
                      s_opt: Optional[ScheduledOptimizer],
                      validation=False) -> Tuple[float, float, float]:
        training = not validation
        model.train(mode=training)
        loss_fn = nn.BCELoss()
        loss_sum = 0
        bp_time = 0
        pass_count = 0
        match_count = 0
        item_count = 0
        data_loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)
        for i in range(self.switch_ds_every):
            for q_tuple, p_tuple, label in data_loader:
                qid, qtl, qam = q_tuple
                pid, pam = p_tuple
                y_pred = model.forward(query_input_ids=qid, query_attention_mask=qam, query_token_lengths=qtl,
                                       passage_input_ids=pid, passage_attention_mask=pam)
                loss = loss_fn(y_pred, label)
                loss_sum += loss.item()
                if training:
                    s_opt.zero_grad()
                    time1 = timer()
                    gc.collect()
                    loss.backward()
                    time2 = timer()
                    bp_time += time2 - time1
                    s_opt.step()
                    s_opt.zero_grad(set_to_none=True)
                else:
                    is_match = (y_pred >= 0.5) == (label >= 0.5)
                    match_count += torch.sum(is_match).item()
                    item_count += label.size(0)
                del y_pred
                del label
                del loss
                gc.collect()
                pass_count += 1
        accuracy = 0 if item_count == 0 else match_count / item_count
        return loss_sum / pass_count, bp_time, accuracy,

    def train(self, model: BaseQueryPassageProbModel):
        print('------------------------------------')
        print(f'ML learning rate: {self.lm_lr}')
        print(f'Extras learning rate: {self.extras_lr}')
        print(f'Batch size: {self.batch_size}')
        print(f'Num epochs per dataset: {self.switch_ds_every}')
        print(f'Num epochs: {self.num_epochs}')
        print(f'Max query tokens: {self.max_query_tokens}')
        print(f'Min passage tokens: {self.min_passage_tokens}')
        print(f'Max passage tokens: {self.max_passage_tokens}')
        print('------------------------------------')
        if self.num_epochs > 0:
            print('Training...')
            lm_parameters = model.get_lm_parameters()
            extra_parameters = model.get_added_parameters()
            num_training_steps = self.num_epochs * math.ceil(self.num_ds_examples / self.batch_size)
            parameters = [
                {'params': lm_parameters, 'lr': self.lm_lr},
                {'params': extra_parameters, 'lr': self.extras_lr},
            ]
            s_opt = ScheduledOptimizer(parameters, lr=self.lm_lr,
                                       num_training_steps=num_training_steps,
                                       num_warmup_steps=self.num_warmup_steps,
                                       weight_decay=self.weight_decay)
            sum_bp_time = 0
            time1 = timer()
            for epoch in range(0, self.num_epochs, self.switch_ds_every):
                dataset = self.create_dataset(training=True)
                train_loss, bp_time, _ = self.train_dataset(dataset, model, s_opt, validation=False)
                sum_bp_time += bp_time
                print(f'-- Epoch {epoch}: Train Loss={train_loss:.4g}')
                last_lrs = s_opt.get_last_lrs()
                print(f'Last LM LR: {np.round(last_lrs, 6)}')
                if self.track_validation and epoch % 20 == 0:
                    valid_ds = self.create_dataset(training=False)
                    valid_loss, _, valid_acc = self.train_dataset(valid_ds, model, None,
                                                                  validation=True)
                    print(f'== Epoch {epoch}: Valid Loss={valid_loss:.4g} | Accuracy={valid_acc:.3f}')
            time2 = timer()
            elapsed = time2 - time1
            print(f'Trained {self.num_epochs} epochs in {elapsed:.1f} seconds.')
            bp_time_pct = sum_bp_time * 100.0 / elapsed
            print(f'Backprop time: {sum_bp_time:.1f} seconds ({bp_time_pct:.1f}%)')
            eph = self.num_epochs / (elapsed / (60 * 60))
            print(f'Training speed: {eph:.1f} epochs per hour.')
            torch.cuda.empty_cache()
            gc.collect()

    def create_dataset(self, training: bool):
        data_sources = self.train_data_sources if training else self.valid_data_sources
        return QueryPassageDataset(data_sources, self.tokenizer, self.num_ds_examples,
                                   device=self.device,
                                   approx_positive_fraction=self.approx_positive_fraction)



import io
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional

from goodai.ltm.mem.chunk import TextKeyType
from goodai.ltm.mem.chunk_queue import PassageInfo


@dataclass
class RetrievedMemory:
    passage: str
    """
    The text of the retrieved passage, or expanded chunk.
    """

    passage_info: PassageInfo
    """
    Additional information about the passage location and tokens.
    """

    timestamp: float
    """
    The timestamp of the memory. By default, this is seconds since Epoch. The user
    may provide custom timestamps when adding text to the memory object.
    """

    distance: float
    """
    A distance metric between the retrieved passage and the query.
    """

    relevance: float
    """
    A metric between 0 and 1 that indicates how relevant to the query 
    the retrieved memory is. It may be derived from the distance or 
    the confidence, if available.
    """

    textKeys: List[TextKeyType]
    """
    Text keys associated with the matching chunk. These are the keys
    returned when add_text() is called.
    """

    confidence: Optional[float] = None
    """
    A confidence metric between 0 and 1. Not all memory implementations 
    support this, so it may be None.
    """

    importance: Optional[float] = None
    """
    An importance score assigned to the retrieved memory by the importance model.
    If an importance model is not available, this property will be None.
    """

    metadata: Optional[dict] = None
    """
    Metadata associated with the retrieved text.
    """

    @staticmethod
    def has_overlap(included_indexes: set[int], p_from: int, p_to: int,
                    overlap_threshold: float):
        if p_from not in included_indexes and (p_to - 1) not in included_indexes:
            return False
        if p_to <= p_from:
            return False
        intersection = included_indexes.intersection(range(p_from, p_to))
        len_inter = len(intersection)
        if len_inter <= 0:
            return False
        overlap_fraction = len_inter / (p_to - p_from)
        return overlap_fraction >= overlap_threshold

    @classmethod
    def remove_overlaps(cls, r_memories: list['RetrievedMemory'],
                        overlap_threshold: float) -> list['RetrievedMemory']:
        result = []
        included_indexes: set[int] = set()
        for item in r_memories:
            p_info = item.passage_info
            if p_info is None:
                continue
            p_from = p_info.fromIndex
            p_to = p_info.toIndex
            if cls.has_overlap(included_indexes, p_from, p_to, overlap_threshold):
                continue
            result.append(item)
            included_indexes.update(range(p_from, p_to))
        return result


class BaseReranker(ABC):
    """
    Abstract base class for custom rerankers.
    """
    @abstractmethod
    def rerank(self, r_memories: List[RetrievedMemory], mem: 'BaseTextMemory') -> List[RetrievedMemory]:
        pass


class BaseImportanceModel(ABC):
    """
    Abstract base class for models that set the importance property of stored memories.
    """

    @abstractmethod
    def get_importance(self, mem_text: str):
        pass


class BaseTextMemory(ABC):
    """
    Abstract base class for text memories.
    """

    @abstractmethod
    def add_text(self, text: str, metadata: Optional[dict] = None, rewrite: bool = False,
                 rewrite_context: Optional[str] = None, show_progress_bar: bool = False,
                 timestamp: Optional[float] = None) -> TextKeyType:
        """
        Adds text to the memory.
        :param show_progress_bar: Whether a progress bar should be shown
        :param text: The string that is appended to the memory
        :param metadata: An optional dictionary with metadata
        :param rewrite: Whether the text should be rewritten by an LLM
        :param rewrite_context: The context provided to the LLM for rewriting the text
        :param timestamp: A custom timestamp for the memory to use instead of time.time()
        :return: A unique key associated with the text that was added to the memory.
        """
        pass

    @abstractmethod
    def replace_text(self, text_key: TextKeyType, text: str, metadata: Optional[dict] = None,
                     rewrite: bool = False, rewrite_context: Optional[str] = None,
                     show_progress_bar: bool = False, timestamp: Optional[float] = None) -> TextKeyType:
        """
        Replaces text stored in the memory.
        :param text_key: A key previously returned by the add_text() method.
        :param text: The new text.
        :param metadata: Any metadata associated with the new text.
        :param rewrite: Whether the text should be rewritten by an LLM.
        :param rewrite_context: The context passed to the LLM to inform the rewrite.
        :param show_progress_bar: Whether a progress bar should be shown.
        :param timestamp: The timestamp associated with the new text.
        :return: The text key.
        """
        pass

    @abstractmethod
    def delete_text(self, text_key: TextKeyType, show_progress_bar: bool = False) -> TextKeyType:
        """
        Deletes text stored in the meory.
        :param text_key: A key previously returned by the add_text() method.
        :param show_progress_bar: Whether a progress bar should be shown.
        :return: The text key.
        """
        pass

    @abstractmethod
    def get_text(self, text_key: TextKeyType) -> Optional[str]:
        """
        :param text_key: A key previously returned by the add_text() method.
        :return: The text associated with the provided key. If the key does not exist, the returned value is None.
        """
        pass

    @abstractmethod
    def add_separator(self):
        """
        Adds a section separator. The last chunk in the memory queue will be filled with padding.
        Chunk expansion cannot go across section boundaries.
        """
        pass

    @abstractmethod
    def is_empty(self) -> bool:
        """
        :return: Whether the memory contains no tokens
        """
        pass

    @abstractmethod
    def retrieve_multiple(self, queries: List[str], k: int, rewrite: bool = False, show_progress_bar: bool = False,
                          **kwargs) -> List[List[RetrievedMemory]]:
        pass

    def retrieve(self, query: str, k: int, rewrite: bool = False, **kwargs) -> List[RetrievedMemory]:
        """
        Performs a memory search and retrieves relevant passages.
        :param query: The query used to search for memories
        :param k: The number of requested memories
        :param rewrite: Whether the query should be rewritten by an LLM
        :param kwargs: Additional argument passed to underlying models
        :return: A list of RetrievedMemory instances.
        """
        multi_result = self.retrieve_multiple([query], k=k, rewrite=rewrite,
                                              **kwargs)
        return multi_result[0]

    @abstractmethod
    def clear(self):
        """
        Clears all content in the memory.
        """
        pass

    @abstractmethod
    def dump(self, stream: io.TextIOBase = sys.stdout):
        """
        Performs a diagnostic dump of the contents of the memory
        :param stream: This IO stream is where text is dumped
        """
        pass

    @abstractmethod
    def has_importance_model(self) -> bool:
        pass



import re
from typing import List, Optional, Any, Dict, Tuple

import numpy as np
from datasets import load_dataset
from transformers import PreTrainedTokenizer

from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource
from goodai.ltm.data.query_passage.example import QueryPassageExample


class WikiQueryPassageDataSource(BaseQueryPassageDataSource):
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer, articles: List[Any],
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 gap_leeway=12, min_query_tokens=16):
        super().__init__()
        if max_query_tokens < min_query_tokens:
            raise ValueError(f'Expected max_query_tokens to be at least {min_query_tokens}')
        self.min_query_tokens = min_query_tokens
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.gap_leeway = gap_leeway
        self.min_par_len = max_query_tokens + max_passage_tokens + gap_leeway
        self.tokenizer = tokenizer
        self.articles = articles
        self.random = random
        self.pos_article_token_ids: Optional[List[List[int]]] = None
        self.neg_article_token_ids: Optional[List[List[int]]] = None
        self.current_par_idx = -1
        self.tokenization_cache: Dict[int, List[List[int]]] = dict()
        self.par_pattern = re.compile(r'\r?\n\r?\n')

    @classmethod
    def create_data_sources(cls, train_fraction: float, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                            max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                            min_article_len: int = 512) ->\
            Tuple['WikiQueryPassageDataSource', 'WikiQueryPassageDataSource']:
        ds = load_dataset('wikipedia', '20220301.simple')
        ds_data = ds['train']
        articles = list(ds_data)
        articles = [a for a in articles if len(a['text']) >= min_article_len]
        random.shuffle(articles)
        train_len = round(len(articles) * train_fraction)
        train_articles = articles[:train_len]
        test_articles = articles[train_len:]
        train_ds = cls(random, tokenizer, train_articles, max_query_tokens, min_passage_tokens, max_passage_tokens)
        test_ds = cls(random, tokenizer, test_articles, max_query_tokens, min_passage_tokens, max_passage_tokens)
        return train_ds, test_ds,

    def get_tokenization(self, index: int) -> List[List[int]]:
        art_token_ids = self.tokenization_cache.get(index)
        if art_token_ids is None:
            index = self.random.randint(0, len(self.articles))
            article = self.articles[index]
            content = article['text']
            paragraphs: List[str] = self.par_pattern.split(content)
            art_token_ids = []
            for p in paragraphs:
                par_token_ids = self.tokenizer.encode(p, add_special_tokens=False)
                if len(par_token_ids) >= self.min_par_len:
                    art_token_ids.append(par_token_ids)
            self.tokenization_cache[index] = art_token_ids
        return art_token_ids

    def is_new_article_needed(self):
        return self.pos_article_token_ids is None or self.current_par_idx >= len(self.pos_article_token_ids)

    def init_current_article(self):
        pos_art_token_ids: List[List[int]] = []
        neg_art_token_ids: List[List[int]] = []
        for attempt in range(100):
            pos_index = self.random.randint(0, len(self.articles))
            pos_art_token_ids = self.get_tokenization(pos_index)
            if len(pos_art_token_ids) == 0:
                continue
            neg_index = self.random.randint(0, len(self.articles))
            if neg_index == pos_index:
                continue
            neg_art_token_ids = self.get_tokenization(neg_index)
            if len(neg_art_token_ids) == 0:
                continue
            break
        if len(pos_art_token_ids) == 0 or len(neg_art_token_ids) == 0:
            raise SystemError('Unable to find article with suitable paragraph lengths.')
        self.neg_article_token_ids = neg_art_token_ids
        self.pos_article_token_ids = pos_art_token_ids
        self.current_par_idx = 0

    def _get_neg_passage_ids(self, num_passage_tokens: int, p_use_different_art: float = 0.9):
        r = self.random
        use_different = r.uniform() < p_use_different_art
        if not use_different:
            from_par_idx = self.current_par_idx + 2
            to_par_idx = len(self.pos_article_token_ids)
            if from_par_idx < to_par_idx:
                neg_par_idx = r.randint(from_par_idx, to_par_idx)
                par_token_ids = self.pos_article_token_ids[neg_par_idx]
                flex = len(par_token_ids) - num_passage_tokens
                if flex >= 0:
                    start = r.randint(0, flex + 1)
                    return par_token_ids[start:start + num_passage_tokens]
        neg_par_id = r.randint(0, len(self.neg_article_token_ids))
        neg_par_token_ids = self.neg_article_token_ids[neg_par_id]
        flex = len(neg_par_token_ids) - num_passage_tokens
        passage_start = r.randint(0, flex + 1)
        passage_end = passage_start + num_passage_tokens
        passage_ids = neg_par_token_ids[passage_start:passage_end]
        return passage_ids

    def sample_items(self, count: int, approx_positive_fraction: float = 0.5) -> List[QueryPassageExample]:
        # Note: sampling not suitable for embedding model training
        r = self.random
        result = []
        for i in range(count):
            if self.is_new_article_needed():
                self.init_current_article()
            par_token_ids = self.pos_article_token_ids[self.current_par_idx]
            num_query_tokens = r.randint(self.min_query_tokens, self.max_query_tokens + 1)
            num_gap_tokens = r.randint(0, self.gap_leeway + 1)
            num_passage_tokens = r.randint(self.min_passage_tokens, self.max_passage_tokens + 1)
            is_match = self.random.uniform() <= approx_positive_fraction
            if is_match:
                total_tokens = num_query_tokens + num_gap_tokens + num_passage_tokens
            else:
                total_tokens = num_query_tokens
            flex = len(par_token_ids) - total_tokens
            start = r.randint(0, flex + 1)
            query_end = start + num_query_tokens
            query_ids = par_token_ids[start:query_end]
            if is_match:
                passage_start = query_end + num_gap_tokens
                passage_end = passage_start + num_passage_tokens
                passage_ids = par_token_ids[passage_start:passage_end]
            else:
                passage_ids = self._get_neg_passage_ids(num_passage_tokens)
            result.append(QueryPassageExample(query_ids, passage_ids, is_match))
            self.current_par_idx += 1
        return result



import itertools
import logging
from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch import nn
from transformers import PreTrainedModel, PreTrainedTokenizer, PretrainedConfig
from transformers.modeling_outputs import BaseModelOutput

from goodai.helpers.tokenizer_helper import get_attention_after_token
from goodai.helpers.torch_helper import param_count
from goodai.ltm.embeddings.trainable import TrainableEmbeddingModel


class DefaultEmbeddingModel(TrainableEmbeddingModel):
    """
    Default implementation of trainable text embeddings.
    """

    def __init__(self, lang_model: PreTrainedModel, tokenizer: PreTrainedTokenizer,
                 num_retrieval_emb: int, num_storage_emb: int,
                 num_end_chars_lb_ignore=18, inner_size=30, dropout=0.03):
        super(DefaultEmbeddingModel, self).__init__(tokenizer)
        lb_token_ids = tokenizer.encode('\n', add_special_tokens=False)
        valid_lb_token_id = len(lb_token_ids) == 1
        if not valid_lb_token_id:
            logging.warning(f'Tokenizer "{tokenizer.name_or_path}" does not have a line-break token.')
        self.lb_token_id = lb_token_ids[0] if valid_lb_token_id else -1
        self.num_end_chars_lb_ignore = num_end_chars_lb_ignore
        lm_config = lang_model.config
        self.lang_model = lang_model
        self.tokenizer = tokenizer
        self.num_storage_emb = num_storage_emb
        self.num_retrieval_emb = num_retrieval_emb
        self.emb_dim = self._get_embed_dim(lang_model, lang_model.config)
        hidden_size = lm_config.hidden_size
        out_size = num_retrieval_emb + num_storage_emb
        self.out_model = nn.Sequential(
            nn.Dropout(p=dropout),
            nn.Linear(hidden_size, inner_size),
            nn.ELU(),
            nn.LayerNorm(inner_size),
            nn.Dropout(p=dropout),
            nn.Linear(inner_size, out_size),
        )
        self.dummy = nn.Parameter()

    @staticmethod
    def _get_embed_dim(model: PreTrainedModel, lm_config: PretrainedConfig):
        return lm_config.hidden_size

    def get_info(self) -> str:
        return f'GoodAI-finetuned model | Base: {self.lang_model.name_or_path} | ' \
               f'Params: {param_count(self)/1e+6:.4g} million | Dimensions: {self.get_embedding_dim()} | ' \
               f'{self.get_num_storage_embeddings()} storage embeddings'

    def get_device(self):
        return self.dummy.device

    def forward(self, is_retrieve: bool, token_lengths: Optional[torch.Tensor], **kwargs):
        input_ids = kwargs['input_ids']
        attention_mask = kwargs.get('attention_mask')
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids)
        if is_retrieve:
            if token_lengths is None:
                raise ValueError('token_lengths is mandatory when requesting retrieval keys')
            after_lb_att_mask = get_attention_after_token(input_ids, attention_mask, token_lengths, self.lb_token_id,
                                                          exclude_last_n_chars=self.num_end_chars_lb_ignore,
                                                          device=self.get_device())
            out_att_mask = attention_mask * after_lb_att_mask
        else:
            if token_lengths is not None:
                raise ValueError('token_lengths parameter was provided but is not expected for storage keys')
            out_att_mask = attention_mask
        lm_output: BaseModelOutput = self.lang_model(input_ids=input_ids,
                                                     attention_mask=attention_mask,
                                                     return_dict=True)
        token_embeddings = lm_output.last_hidden_state
        # token_embeddings: (batch_size, seq_len, emb_size,)
        slf_att_logits_1 = self.out_model(token_embeddings)
        slf_att_logits = slf_att_logits_1 - 200 * (1.0 - out_att_mask[:, :, None])
        slf_att_logits = torch.clamp(slf_att_logits, min=-100, max=+80)
        slf_att_weights = torch.softmax(slf_att_logits, dim=1)
        # slf_att_weights: (batch_size, seq_len, total_keys,)
        emb_product = token_embeddings[:, :, None, :] * slf_att_weights[:, :, :, None]
        # emb_product: (batch_size, seq_len, total_keys, emb_size,)
        if is_retrieve:
            relevant_product = emb_product[:, :, self.num_storage_emb:, :]
        else:
            relevant_product = emb_product[:, :, :self.num_storage_emb, :]
        raw_key_mean = torch.sum(relevant_product, dim=1)
        # raw_key_mean: (batch_size, num_keys, emb_size,)
        return F.normalize(raw_key_mean, dim=2)

    def get_lm_parameters(self):
        return self.lang_model.parameters()

    def get_added_parameters(self):
        return itertools.chain(self.out_model.parameters())

    def get_storage_emb(self, input_ids: torch.Tensor, attention_mask: torch.Tensor = None) -> torch.Tensor:
        return self(input_ids=input_ids, token_lengths=None, attention_mask=attention_mask, is_retrieve=False)

    def get_retrieval_emb(self, input_ids: torch.Tensor, token_lengths: torch.Tensor,
                          attention_mask: torch.Tensor = None) -> torch.Tensor:
        return self(input_ids=input_ids, token_lengths=token_lengths, attention_mask=attention_mask,
                    is_retrieve=True)

    def get_emb(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        sk = self.get_storage_emb(input_ids, attention_mask)
        rk = self.get_retrieval_emb(input_ids, attention_mask)
        return sk, rk,

    def get_embedding_dim(self) -> int:
        return self.emb_dim

    def get_num_retrieval_embeddings(self) -> int:
        return self.num_retrieval_emb

    def get_num_storage_embeddings(self) -> int:
        return self.num_storage_emb



import enum
import sys
import weakref

import torch
from typing import Union, Optional
from transformers import AutoTokenizer, PreTrainedTokenizer
from goodai.ltm.embeddings.auto import AutoTextEmbeddingModel
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel
from goodai.ltm.mem.base import BaseTextMemory, BaseReranker, BaseImportanceModel
from goodai.ltm.mem.chunk_queue import ChunkQueue
from goodai.ltm.mem.config import TextMemoryConfig
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.mem.mem_foundation import VectorDbType
from goodai.ltm.mem.rewrite_model import BaseRewriteModel
from goodai.ltm.mem.simple_vector_db import SimpleVectorDb
from goodai.ltm.reranking.auto import AutoTextMatchingModel
from goodai.ltm.reranking.base import BaseTextMatchingModel

_default_tokenizer_wr: Optional[weakref.ref] = None


class MemType(enum.Enum):
    TRANSIENT_CHUNK_EMB = 0


class AutoTextMemory:
    """
    Factory class for text memory.
    """

    @staticmethod
    def create(mem_type: MemType = MemType.TRANSIENT_CHUNK_EMB,
               vector_db_type: VectorDbType = VectorDbType.SIMPLE,
               tokenizer: PreTrainedTokenizer = None,
               emb_model: Union[BaseTextEmbeddingModel, Optional[str]] = None,
               matching_model: Union[BaseTextMatchingModel, Optional[str]] = None,
               query_rewrite_model: BaseRewriteModel = None,
               memory_rewrite_model: BaseRewriteModel = None,
               reranker: BaseReranker = None,
               importance_model: BaseImportanceModel = None,
               device: Union[torch.device, str] = None,
               config: TextMemoryConfig = None,
               chunk_queue: Optional[ChunkQueue] = None,
               vector_db: Optional[SimpleVectorDb] = None,
               ) -> BaseTextMemory:
        """
        Creates a memory instance.
        :param mem_type: Reserved parameter.
        :param vector_db_type: The type of vector database. Default is VectorDbType.SIMPLE.
        :param tokenizer: A chunking tokenizer. It should be a tokenizer that preserves whitespace and casing.
        :param emb_model: The embedding model.
        :param matching_model: An optional query-passage matching model.
        :param query_rewrite_model: The query rewrite model.
        :param memory_rewrite_model: The memory rewrite model.
        :param reranker: A custom reranker.
        :param importance_model: A model that assigns an importance value to stored memories.
        It may be required by some rerankers.
        :param device: The Pytorch device.
        :param config: The memory configuration.
        :param chunk_queue: The chunk queue, only when loading the memory.
        :param vector_db: The vector db, only when loading the memory.
        :return: An instance of BaseTextMemory.
        """
        if tokenizer is None:
            global _default_tokenizer_wr

            if _default_tokenizer_wr:
                tokenizer = _default_tokenizer_wr()
            if tokenizer is None:
                tokenizer = AutoTokenizer.from_pretrained('distilroberta-base')
                # Suppress length warning
                tokenizer.model_max_length = sys.maxsize
                _default_tokenizer_wr = weakref.ref(tokenizer)
        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
        if emb_model is None:
            emb_model = 'em-distilroberta-p1-01'
        if isinstance(emb_model, str):
            emb_model = AutoTextEmbeddingModel.shared_pretrained(emb_model)
        if isinstance(matching_model, str):
            matching_model = AutoTextMatchingModel.shared_pretrained(matching_model)
        if config is None:
            config = TextMemoryConfig()
        return DefaultTextMemory(vector_db_type, tokenizer, emb_model, matching_model,
                                 device=device, config=config,
                                 query_rewrite_model=query_rewrite_model,
                                 memory_rewrite_model=memory_rewrite_model,
                                 reranker=reranker,
                                 importance_model=importance_model,
                                 chunk_queue=chunk_queue,
                                 vector_db=vector_db,
                                 )



import pickle
from typing import Union
from weakref import WeakValueDictionary

import torch

from goodai.helpers.file_helper import open_url_as_file, unpickle_downloaded_url
from goodai.ltm.embeddings.openai_emb import OpenAIEmbeddingModel
from goodai.ltm.embeddings.st_emb import SentenceTransformerEmbeddingModel
from goodai.ltm.embeddings.trainable import TrainableEmbeddingModel
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel

_models_base = 'https://github.com/GoodAI/goodai-ltm-artifacts/releases/download/models/goodai-ltm-emb-model'

_pretrained_map = {
    'em-distilroberta-p1-01': f'{_models_base}-p1-1118',
    'em-distilroberta-p3-01': f'{_models_base}-p3-1116',
    'em-MiniLM-p3-01': f'{_models_base}-p3-1117',
    'em-MiniLM-p1-01': f'{_models_base}-p1-1151',
    'em-distilroberta-p5-01': f'{_models_base}-p5-1150',
}


class AutoTextEmbeddingModel:
    """
    Factory class for text embedding models.
    """

    @staticmethod
    def shared_pretrained(name: str, device: Union[str, torch.device] = None):
        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
        key = f'{name}|{device}'
        model = AutoTextEmbeddingModel.model_cache.get(key)
        if model is None:
            model = AutoTextEmbeddingModel.from_pretrained(name, device)
            AutoTextEmbeddingModel.model_cache[key] = model
        return model

    @staticmethod
    def from_pretrained(name: str, device: Union[str, torch.device] = None, **kwargs) -> BaseTextEmbeddingModel:
        """
        Makes a pretrained embedding model from a descriptor (name).

        The name has the format <model_type>:<model_name>, for example "st:sentence-transformers/all-distilroberta-v1",
        where model_type is 'st' for Hugging Face Sentence Transformers or 'openai' for Open AI text embeddings.

        :param name: Name in the format <model_type>:<model_name>
        :param device: The Pytorch device for the model, if applicable
        :return: The embedding model
        """

        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
        name = name.strip()
        colon_idx = name.find(':')
        if colon_idx == -1:
            url = _pretrained_map.get(name)
            if url is None:
                raise ValueError(f'GoodAI model not found: {name}')
            model_dict = unpickle_downloaded_url(url)
            model: TrainableEmbeddingModel = model_dict['emb_model']
            model.to(device)
            model.zero_grad(set_to_none=True)
            model.eval()
            return model
        model_type = name[:colon_idx]
        model_name = name[colon_idx + 1:]
        if model_type == 'st':
            return SentenceTransformerEmbeddingModel(model_name, device=device, **kwargs)
        elif model_type == 'openai':
            return OpenAIEmbeddingModel(model_name, device=device, **kwargs)
        else:
            raise ValueError(f'Unknown model type: {model_type}')


AutoTextEmbeddingModel.model_cache = WeakValueDictionary()



import json
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Callable

import numpy

from goodai.ltm.mem.auto import AutoTextMemory
from goodai.ltm.mem.chunk import Chunk
from goodai.ltm.mem.chunk_queue import ChunkQueue
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.mem.simple_vector_db import SimpleVectorDb
from goodai.ltm.persistence.abstract import MemoryPersistence


def default_obj_to_dict(obj):
    if isinstance(obj, numpy.ndarray):
        return obj.tolist()
    if isinstance(obj, datetime):
        return {"type": "datetime", "value": obj.isoformat()}
    if hasattr(obj, "__dict__"):
        obj_dict = obj.__dict__.copy()
        obj_dict.pop("chunk_map", None)
        return obj_dict
    raise TypeError(f"{type(obj)} is not JSON/Msgpack serializable")


def default_dict_to_obj(dct):
    if "type" in dct:
        if dct["type"] == "datetime":
            return datetime.fromisoformat(dct["value"])
    return dct


def serialize_to_json(obj: Any, obj_to_dict: Callable) -> str:
    return json.dumps(obj, default=obj_to_dict)


def save_vector_db_to_npz(vector_db: SimpleVectorDb, file: Path):
    numpy.savez(file, all_vectors=vector_db.all_vectors, all_ids=vector_db.all_ids)


def load_vector_db_from_npz(file: Path) -> SimpleVectorDb:
    with numpy.load(file) as data:
        all_vectors = data["all_vectors"]
        all_ids = data["all_ids"]
        vector_db = SimpleVectorDb(all_vectors=all_vectors, all_ids=all_ids)
    return vector_db


def save_vector_db_to_json(vector_db: SimpleVectorDb, file: Path, obj_to_dict: Callable = default_obj_to_dict):
    with open(file, "w", encoding="utf-8") as fd:
        fd.write(serialize_to_json(vector_db, obj_to_dict))


def load_vector_db_from_json(file: Path) -> SimpleVectorDb:
    with open(file, "r", encoding="utf-8") as fd:
        return load_vector_db_from_dict(json.load(fd))


def load_vector_db_from_dict(loaded_vector_db: Dict) -> SimpleVectorDb:
    return SimpleVectorDb(
        all_vectors=numpy.array(loaded_vector_db["all_vectors"]),
        all_ids=numpy.array(loaded_vector_db["all_ids"])
    )


def load_chunk_queue_from_dict(loaded_chunk_queue: Dict) -> ChunkQueue:
    typed_chunks = []
    for chunk in loaded_chunk_queue["chunks"]:
        c = Chunk(**chunk)
        typed_chunks.append(
            c
        )
    sequence_map = loaded_chunk_queue["sequence_map"]
    # json and msgpack don't allow string keys so it has to be converted back
    # values have to be converted from list to tuple
    loaded_chunk_queue["sequence_map"] = {int(key): tuple(value) for key, value in sequence_map.items()}
    loaded_chunk_queue["chunks"] = typed_chunks
    chunk_queue = ChunkQueue(**loaded_chunk_queue)
    return chunk_queue


@dataclass
class JsonMemoryPersistence(MemoryPersistence):
    """
    Saves and loads the memory to json files.
    Space inefficient since floats are stored as ascii characters.
    """
    chunk_queue_filename: str = "chunk_queue.json"
    vector_db_npz_filename = "vector_db.json"
    obj_to_dict: Callable = default_obj_to_dict
    dict_to_obj: Callable = default_dict_to_obj

    def save(self, memory: DefaultTextMemory, directory: Path):
        chunk_queue = memory.chunk_queue
        file = directory / self.chunk_queue_filename
        with open(file, "w", encoding="utf-8") as fd:
            fd.write(serialize_to_json(chunk_queue, self.obj_to_dict))

        vector_db = memory.vector_db
        file = directory / self.vector_db_npz_filename
        save_vector_db_to_json(vector_db, file)

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.chunk_queue_filename
        chunk_queue = load_chunk_queue_from_dict(
            json.load(open(file, "r", encoding="utf-8"), object_hook=self.dict_to_obj)
        )

        file = directory / self.vector_db_npz_filename
        vector_db = load_vector_db_from_json(file)
        return AutoTextMemory.create(chunk_queue=chunk_queue, vector_db=vector_db, **kwargs)

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.chunk_queue_filename,
            directory / self.vector_db_npz_filename,
        ]
        return all([file.exists() for file in files])


@dataclass
class JsonNpzMemoryPersistence(MemoryPersistence):
    """
    Saves chunk queue to json and vector db to npz, which is space efficient.
    """
    chunk_queue_filename: str = "chunk_queue.json"
    vector_db_npz_filename = "vector_db.npz"
    obj_to_dict: Callable = default_obj_to_dict
    dict_to_obj: Callable = default_dict_to_obj

    def save(self, memory: DefaultTextMemory, directory: Path):
        chunk_queue = memory.chunk_queue
        file = directory / self.chunk_queue_filename
        with open(file, "w", encoding="utf-8") as fd:
            fd.write(serialize_to_json(chunk_queue, self.obj_to_dict))

        vector_db = memory.vector_db
        file = directory / self.vector_db_npz_filename
        save_vector_db_to_npz(vector_db, file)

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.chunk_queue_filename
        chunk_queue = load_chunk_queue_from_dict(
            json.load(open(file, "r", encoding="utf-8"), object_hook=self.dict_to_obj))

        file = directory / self.vector_db_npz_filename
        vector_db = load_vector_db_from_npz(file)

        return AutoTextMemory.create(
            chunk_queue=chunk_queue, vector_db=vector_db,
            **kwargs
        )

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.chunk_queue_filename,
            directory / self.vector_db_npz_filename,
        ]
        return all([file.exists() for file in files])



import codecs
import json
import os
from typing import List

import numpy as np
from transformers import PreTrainedTokenizer

from goodai.helpers.file_helper import download_zip
from goodai.ltm.eval.mem import BaseMemEvaluator, QAScenario

_url = 'https://github.com/apple/ml-qrecc/blob/main/dataset/qrecc_data.zip?raw=true'
_test_file = 'qrecc_test.json'


class QreccMemEvaluator(BaseMemEvaluator):
    def __init__(self, tokenizer: PreTrainedTokenizer, top_ks: List[int],
                 max_query_tokens: int, has_query_noise: bool, use_rewrite: bool = True,
                 max_scenarios: int = 2000):
        super().__init__(tokenizer, top_ks, max_query_tokens, has_query_noise)
        self.use_rewrite = use_rewrite
        data_dir = download_zip(_url)
        data_file = os.path.join(data_dir, _test_file)
        with codecs.open(data_file, 'r', 'utf-8') as fd:
            all_data = json.load(fd)
            rnd = np.random.RandomState(8009)
            rnd.shuffle(all_data)
            self.data = all_data[:max_scenarios]

    def get_facts_to_be_stored(self) -> List[str]:
        facts = []
        for entry in self.data:
            answer = entry.get('Answer')
            if answer:
                facts.append(answer)
        return facts

    def get_scenarios(self) -> List[QAScenario]:
        scenarios = []
        for entry in self.data:
            answer = entry.get('Answer')
            if answer:
                context = entry.get('Context')
                if context is not None:
                    q = entry.get('Rewrite') if self.use_rewrite else entry.get('Question')
                    if q:
                        sf = [answer]
                        scenario = QAScenario(context=context, supportingFacts=sf, question=q)
                        scenarios.append(scenario)
        return scenarios



from abc import ABC, abstractmethod
from typing import List, Tuple


class BaseTextMatchingModel(ABC):
    """
    Abstract base class for text matching models.

    A matching model measures the match between queries and passages.
    """

    def get_match_confidence(self, query: str, passages: List[str], batch_size: int = 32,
                             show_progress_bar: bool = False) -> List[float]:
        """
        :param query: A string representing a query.
        :param passages: A list of strings representing passages that should be matched with the provided query.
        :param show_progress_bar: Whether a progress bar should be shown.
        :param batch_size: The inference batch size.
        :return: A list of probabilities or confidence scores (0 to 1) quantifying query-passage matches.
        """
        sentences = [(query, p) for p in passages]
        return self.predict(sentences, batch_size=batch_size, show_progress_bar=show_progress_bar)

    @abstractmethod
    def predict(self, sentences: List[Tuple[str, str]], batch_size: int = 32,
                show_progress_bar: bool = False) -> List[float]:
        pass

    @abstractmethod
    def get_info(self):
        pass



from typing import Tuple

import numpy as np
from goodai.ltm.data.query_passage.qa import QAQueryPassageDataSource

from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource
from transformers import PreTrainedTokenizer

from goodai.ltm.data.query_passage.sharc import SharcQueryPassageDataSource
from goodai.ltm.data.query_passage.wiki import WikiQueryPassageDataSource
from goodai.ltm.data.query_passage.wikianswers import WikiAnswersQueryPassageDataSource


class AutoQueryPassageDataSource:
    """
    Factory class for training and evaluation data sources for query-passage matching.
    """

    def __init__(self):
        pass

    @staticmethod
    def create(name: str, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
               max_query_tokens=40, min_passage_tokens=24,
               max_passage_tokens=36) -> Tuple[BaseQueryPassageDataSource, BaseQueryPassageDataSource]:
        """
        :param name: The dataset name
        :param random: An instance of a numpy RandomState object
        :param tokenizer: A Huggingface tokenizer
        :param max_query_tokens: The maximum number of query tokens
        :param min_passage_tokens: The minimum number of passage tokens
        :param max_passage_tokens: The maximum number of passage tokens
        :return: Tuple of train and test data sources
        """
        common_params = dict(
            random=random, tokenizer=tokenizer,
            max_query_tokens=max_query_tokens,
            min_passage_tokens=min_passage_tokens,
            max_passage_tokens=max_passage_tokens
        )
        if name in ['coqa', 'squad_v2', 'adversarial_qa']:
            return QAQueryPassageDataSource.create_data_sources(ds_name=name, **common_params)
        elif name == 'sharc':
            return SharcQueryPassageDataSource.create_data_sources(ds_name=name, **common_params)
        elif name == 'wiki':
            return WikiQueryPassageDataSource.create_data_sources(train_fraction=0.8, **common_params)
        elif name == 'wikianswers':
            return WikiAnswersQueryPassageDataSource.create_data_sources(train_fraction=0.8, **common_params)
        else:
            raise ValueError(f'Unknown: {name}')



import gc
import unittest
from typing import List, Tuple, cast

from transformers import AutoTokenizer

from goodai.ltm.embeddings.auto import AutoTextEmbeddingModel
from goodai.ltm.eval.metrics import get_correctness_score
from goodai.ltm.mem.auto import AutoTextMemory
from goodai.ltm.mem.base import BaseReranker, RetrievedMemory, BaseTextMemory, BaseImportanceModel
from goodai.ltm.mem.config import TextMemoryConfig, ChunkExpansionConfig, ChunkExpansionLimitType
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.reranking.base import BaseTextMatchingModel


class TestMem(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._lr_emb_model = AutoTextEmbeddingModel.shared_pretrained('em-MiniLM-p1-01')
        cls._text = "Earth has a dynamic atmosphere, which sustains Earth's surface conditions and protects " \
                    "it from most meteoroids and UV-light at entry. It has a composition of primarily nitrogen " \
                    "and oxygen. Water vapor is widely present in the atmosphere, forming clouds that cover most " \
                    "of the planet. The water vapor acts as a greenhouse gas and, together with other greenhouse " \
                    "gases in the atmosphere, particularly carbon dioxide (CO2), creates the conditions for both " \
                    "liquid surface water and water vapor to persist via the capturing of energy from the Sun's " \
                    "light. This process maintains the current average surface temperature of 14.76 °C, at " \
                    "which water is liquid under atmospheric pressure. Differences in the amount of captured energy " \
                    "between geographic regions (as with the equatorial region receiving more sunlight than the " \
                    "polar regions) drive atmospheric and ocean currents, producing a global climate system with " \
                    "different climate regions, and a range of weather phenomena such as precipitation, allowing " \
                    "components such as nitrogen to cycle."

    def test_basic_usage(self):
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model)
        mem.add_text(self._text, metadata={'foo': 'bar'})
        query = "Is water vapor widely present in the atmosphere?"
        r_memories = mem.retrieve(query, k=2)
        r_combined_text = '\n'.join([rm.passage for rm in r_memories])
        assert "Water vapor is widely present in the atmosphere" in r_combined_text
        m = r_memories[0].metadata
        assert m is not None and m.get('foo') == 'bar'

    def test_no_redundancy(self):
        config = TextMemoryConfig()
        config.redundancy_overlap_threshold = 0.5
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        mem.add_text(self._text, metadata={'foo': 'bar'})
        query = "Is water vapor widely present in the atmosphere? What forms clouds all over the planet?"
        r_memories = mem.retrieve(query, k=3)
        tokenizer = AutoTokenizer.from_pretrained('distilroberta-base')
        for i in range(len(r_memories)):
            p1 = r_memories[i].passage
            for j in range(i + 1, len(r_memories)):
                p2 = r_memories[j].passage
                cs = get_correctness_score(tokenizer, p1, p2)
                assert cs <= 40

    def test_redundancy_allowed(self):
        config = TextMemoryConfig()
        config.redundancy_overlap_threshold = 1.0
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        mem.add_text(self._text, metadata={'foo': 'bar'})
        query = "Is water vapor widely present in the atmosphere? What forms clouds all over the planet?"
        r_memories = mem.retrieve(query, k=3)
        tokenizer = AutoTokenizer.from_pretrained('distilroberta-base')
        match_count = 0
        for i in range(len(r_memories)):
            p1 = r_memories[i].passage
            for j in range(i + 1, len(r_memories)):
                p2 = r_memories[j].passage
                cs = get_correctness_score(tokenizer, p1, p2)
                if cs >= 50:
                    match_count += 1
        assert match_count > 0

    def test_multi_creation(self):
        # Should be able to create many default instances without running out of memory
        mems = []
        for i in range(100):
            mems.append(AutoTextMemory.create())

    def test_multi_creation_with_name(self):
        # Should be able to create many default instances without running out of memory
        mems = []
        for i in range(100):
            mems.append(AutoTextMemory.create(emb_model='em-MiniLM-p1-01'))

    def test_multi_creation_with_qpm_model(self):
        # Should be able to create many default instances without running out of memory
        mems = []
        for i in range(100):
            mems.append(AutoTextMemory.create(emb_model='em-MiniLM-p1-01',
                                              matching_model='em:em-MiniLM-p1-01'))

    def test_create_after_delete(self):
        mem1 = AutoTextMemory.create()
        mem1.add_text('foobar')
        del mem1
        gc.collect()
        mem2 = AutoTextMemory.create()
        mem2.add_text('foobar')

    def test_blank_query(self):
        mem1 = AutoTextMemory.create()
        mem1.add_text('foobar')
        mem1.retrieve('', k=5)

    def test_nearly_empty_mem(self):
        mem1 = AutoTextMemory.create()
        mem1.add_text('a')
        r_memories = mem1.retrieve('Is there something?', k=5)
        assert len(r_memories) > 0

    def test_custom_qpm_model(self):
        class _CustomQPM(BaseTextMatchingModel):
            def predict(self, sentences: List[Tuple[str, str]], batch_size: int = 32,
                        show_progress_bar: bool = False) -> List[float]:
                scores = []
                for query, passage in sentences:
                    if 'UV-light' in passage:
                        score = 0.9
                    elif 'greenhouse' in passage:
                        score = 0.7
                    else:
                        score = 0.5
                    scores.append(score)
                return scores

            def get_info(self):
                return ''

        qpm = _CustomQPM()
        mem = AutoTextMemory.create(matching_model=qpm, emb_model=self._lr_emb_model)
        mem.add_text(self._text)
        r_memories = mem.retrieve('Is water vapor widely present in the atmosphere?', k=5)
        assert 'UV-light' in r_memories[0].passage
        assert 'greenhouse' in r_memories[1].passage

    def test_reranking_count_1(self):
        s_counts = []

        class _LocalReranker(BaseReranker):
            def rerank(self, _r_memories: List[RetrievedMemory], _mem: BaseTextMemory) -> List[RetrievedMemory]:
                nonlocal s_counts
                s_counts.append(len(_r_memories))
                result = list(_r_memories)
                result.sort(key=lambda _m: _m.relevance)
                return result

        config = TextMemoryConfig()
        config.reranking_k_factor = 7
        config.chunk_capacity = 8
        config.redundancy_overlap_threshold = 0.5
        config.chunk_expansion_config = ChunkExpansionConfig.for_chunk()
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, reranker=_LocalReranker(),
                                    config=config)
        mem.add_text(self._text)
        mem.retrieve('Is water vapor widely present in the atmosphere?', k=3)
        self.assertEqual([21], s_counts)

    def test_separators_and_replacement(self):
        facts = [
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
            'Darth Vader is portrayed as a man who always appears in black full-body armor and a mask.',
            'Tony Bennett had four children.'
        ]
        config = TextMemoryConfig()
        config.chunk_capacity = 128
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        text_keys = []
        for i, fact in enumerate(facts):
            text_key = mem.add_text(fact, metadata={'index': i}, timestamp=i + 5)
            text_keys.append(text_key)
            mem.add_separator()
        facts[1] = "Higher education, also called post-secondary education, third-level or " \
                   "tertiary education, is an optional final stage of formal learning that " \
                   "occurs after completion of secondary education."
        mem.replace_text(text_keys[1], facts[1], metadata={'replacement': True}, timestamp=1 + 5)
        is_replacement = [False] * len(facts)
        is_replacement[1] = True
        for i, query in enumerate(facts):
            r_memories = mem.retrieve(query, k=1)
            self.assertEqual(query.strip(), r_memories[0].passage.strip())
            self.assertAlmostEqual(i + 5, r_memories[0].timestamp)
            if is_replacement[i]:
                self.assertTrue(r_memories[0].metadata['replacement'])
            else:
                self.assertEqual(i, r_memories[0].metadata['index'])

    def test_chunk_indexing_with_separators(self):
        facts = [
            "Higher education, also called post-secondary education, third-level or "
            "tertiary education, is an optional final stage of formal learning that "
            "occurs after completion of secondary education.",
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
            'Darth Vader is portrayed as a man who always appears in black full-body armor and a mask.',
            'Tony Bennett had four children.'
        ]
        config = TextMemoryConfig()
        config.chunk_capacity = 8
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        for i, fact in enumerate(facts):
            mem.add_text(fact)
            mem.add_separator()
        mem_default: DefaultTextMemory = cast(DefaultTextMemory, mem)
        chunks, _ = mem_default.chunk_queue.get_chunks_for_indexing()
        assert len(chunks) <= 3

    def test_delete_all(self):
        facts = [
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
            'Darth Vader is portrayed as a man who always appears in black full-body armor and a mask.',
            'Tony Bennett had four children.'
        ]
        config = TextMemoryConfig()
        config.chunk_capacity = 128
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        text_keys = []
        for i, fact in enumerate(facts):
            text_key = mem.add_text(fact, metadata={'index': i}, timestamp=i + 5)
            text_keys.append(text_key)
            mem.add_separator()
        for tk in text_keys:
            mem.delete_text(tk)
        assert mem.is_empty()

    def test_expansion_to_paragraph(self):
        _text = "Earth has a dynamic atmosphere, which sustains Earth's surface conditions and protects " \
                "it from most meteoroids and UV-light at entry. It has a composition of primarily nitrogen " \
                "and oxygen. Water vapor is widely present in the atmosphere, forming clouds that cover most " \
                "of the planet. The water vapor acts as a greenhouse gas and, together with other greenhouse " \
                "gases in the atmosphere, particularly carbon dioxide (CO2), creates the conditions for both " \
                "liquid surface water and water vapor to persist via the capturing of energy from the Sun's " \
                "light.\n\n" \
                "This process maintains the current average surface temperature of 14.76 °C, at " \
                "which water is liquid under atmospheric pressure. Differences in the amount of captured energy " \
                "between geographic regions (as with the equatorial region receiving more sunlight than the " \
                "polar regions) drive atmospheric and ocean currents, producing a global climate system with " \
                "different climate regions, and a range of weather phenomena such as precipitation, allowing " \
                "components such as nitrogen to cycle."

        config = TextMemoryConfig()
        config.chunk_expansion_config = ChunkExpansionConfig.for_paragraph()
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        mem.add_text(_text)
        r1 = mem.retrieve("What is the composition of Earth's atmosphere?", k=1)[0]
        self.assertTrue(r1.passage.strip().startswith('Earth has a dynamic atmosphere'))
        self.assertTrue(r1.passage.strip().endswith("from the Sun's light."))
        r2 = mem.retrieve("What drives atmospheric and ocean currents?", k=1)[0]
        self.assertTrue(r2.passage.strip().startswith('This process maintains'))
        self.assertTrue(r2.passage.strip().endswith("such as nitrogen to cycle."))

    def test_expansion_to_lines(self):
        _text = "Earth has a dynamic atmosphere, which sustains Earth's surface conditions and protects " \
                "it from most meteoroids and UV-light at entry. It has a composition of primarily nitrogen " \
                "and oxygen.\n" \
                "Water vapor is widely present in the atmosphere, forming clouds that cover most " \
                "of the planet. The water vapor acts as a greenhouse gas and, together with other greenhouse " \
                "gases in the atmosphere, particularly carbon dioxide (CO2), creates the conditions for both " \
                "liquid surface water and water vapor to persist via the capturing of energy from the Sun's " \
                "light.\n" \
                "This process maintains the current average surface temperature of 14.76 °C, at " \
                "which water is liquid under atmospheric pressure. Differences in the amount of captured energy " \
                "between geographic regions (as with the equatorial region receiving more sunlight than the " \
                "polar regions) drive atmospheric and ocean currents, producing a global climate system with " \
                "different climate regions, and a range of weather phenomena such as precipitation,\n" \
                "allowing components such as nitrogen to cycle."

        config = TextMemoryConfig()
        config.chunk_expansion_config = ChunkExpansionConfig.for_line_break()
        config.chunk_expansion_config.min_extra_side_tokens = 0
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        mem.add_text(_text)
        r1 = mem.retrieve("Other than water vapor, what are other greenhouse gases?", k=1)[0]
        self.assertTrue(r1.passage.strip().startswith('Water vapor is widely present'))
        self.assertTrue(r1.passage.strip().endswith("from the Sun's light."))
        r2 = mem.retrieve("What drives atmospheric and ocean currents?", k=1)[0]
        self.assertTrue(r2.passage.strip().startswith('This process maintains'))
        self.assertTrue(r2.passage.strip().endswith("phenomena such as precipitation,"))

    def test_expansion_to_sections(self):
        config = TextMemoryConfig()
        config.chunk_expansion_config = ChunkExpansionConfig.for_section()
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)
        facts = [
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
        ]
        mem.add_text(facts[0])
        mem.add_separator()
        mem.add_text(self._text)
        mem.add_separator()
        mem.add_text(facts[1])
        r1 = mem.retrieve("Other than water vapor, what are other greenhouse gases?", k=1)[0]
        self.assertTrue(r1.passage.strip().startswith('Earth has a dynamic atmosphere'))
        self.assertTrue(r1.passage.strip().endswith("components such as nitrogen to cycle."))

    def test_excessive_chunk_expansion(self):
        cec = ChunkExpansionConfig(2048, limit_type=ChunkExpansionLimitType.SECTION)
        config = TextMemoryConfig()
        config.chunk_capacity = 12
        config.chunk_overlap_fraction = 0.5
        config.redundancy_overlap_threshold = 0.5
        config.chunk_expansion_config = cec
        with self.assertRaises(ValueError):
            AutoTextMemory.create(emb_model=self._lr_emb_model, config=config)

    def test_custom_reranker(self):
        class _LocalReranker(BaseReranker):
            def rerank(self, _r_memories: List[RetrievedMemory], _mem: BaseTextMemory) -> List[RetrievedMemory]:
                result = list(_r_memories)
                result.sort(key=lambda _m: _m.relevance)
                return result

        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, reranker=_LocalReranker())
        mem.add_text(self._text)
        r_memories = mem.retrieve('Is water vapor widely present in the atmosphere?', k=5)
        self.assertIn('Water vapor is widely present', r_memories[-1].passage.strip())

    def test_custom_importance_model(self):
        class _LocalImportanceModel(BaseImportanceModel):
            def get_importance(self, mem_text: str):
                if 'kayaks' in mem_text.lower():
                    return 0.25
                elif 'vader' in mem_text.lower():
                    return 0.50
                else:
                    return 0

        facts = [
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
            'Darth Vader is portrayed as a man who always appears in black full-body armor and a mask.',
            'Tony Bennett had four children.'
        ]
        config = TextMemoryConfig()
        config.chunk_capacity = 128
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model, importance_model=_LocalImportanceModel(),
                                    config=config)
        for i, fact in enumerate(facts):
            mem.add_text(fact, metadata={'index': i})
            mem.add_separator()

        r_memories = mem.retrieve('Generic question about anything.', k=5)
        for m in r_memories:
            p = m.passage.lower()
            if 'kayaks' in p:
                self.assertAlmostEqual(m.importance, 0.25)
            elif 'vader' in p:
                self.assertAlmostEqual(m.importance, 0.50)
            else:
                self.assertAlmostEqual(m.importance, 0)

    def test_get_text(self):
        facts = [
            'Cane toads have a life expectancy of 10 to 15 years in the wild.',
            'Kayaks are used to transport people in water.',
            'Darth Vader is portrayed as a man who always appears in black full-body armor and a mask.',
            'Tony Bennett had four children.'
        ]
        config = TextMemoryConfig()
        mem = AutoTextMemory.create(emb_model=self._lr_emb_model,
                                    config=config)
        text_keys = []
        for i, fact in enumerate(facts):
            tk = mem.add_text(fact, metadata={'index': i})
            text_keys.append(tk)

        for tk, fact in zip(text_keys, facts):
            text = mem.get_text(tk)
            self.assertEqual(fact, text)

        text = mem.get_text(99999)
        self.assertIsNone(text)




from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any

TextKeyType = int


@dataclass
class Chunk:
    chunk_id: int
    capacity: int
    from_token_seq_id: int
    metadata: Optional[Dict[Any, Any]]
    importance: Optional[float]
    timestamp: float
    to_token_seq_id: int = None
    indexed_length: int = field(default=-1)
    associated_keys: List[TextKeyType] = field(default_factory=list)

    def __post_init__(self):
        if not self.to_token_seq_id:
            self.to_token_seq_id = self.from_token_seq_id

    def __len__(self):
        return self.to_token_seq_id - self.from_token_seq_id

    def get_room(self):
        return self.capacity - len(self)

    def is_at_capacity(self) -> bool:
        return len(self) >= self.capacity

    def set_to_token_seq_id(self, to_token_seq_id: int):
        if to_token_seq_id - self.from_token_seq_id > self.capacity:
            raise ValueError(f'capacity={self.capacity} would be exceeded')
        self.to_token_seq_id = to_token_seq_id

    def extend_by(self, num_tokens: int):
        if len(self) + num_tokens > self.capacity:
            raise ValueError(f'capacity={self.capacity} would be exceeded')
        self.to_token_seq_id += num_tokens

    def shift(self, offset: int):
        self.from_token_seq_id += offset
        self.to_token_seq_id += offset

    def is_indexed(self) -> bool:
        return self.indexed_length == len(self)

    def update_indexed_state(self):
        self.indexed_length = len(self)

    def add_key(self, text_key: TextKeyType):
        self.associated_keys.append(text_key)



import codecs
import json
import os
from typing import List

import numpy as np
from transformers import PreTrainedTokenizer

from goodai.helpers.file_helper import download_zip
from goodai.ltm.eval.mem import BaseMemEvaluator, QAScenario

_url = 'https://storage.googleapis.com/ai2i/strategyqa/data/strategyqa_dataset.zip'
_test_file = 'strategyqa_train.json'


class StrategyQAMemEvaluator(BaseMemEvaluator):
    def __init__(self, tokenizer: PreTrainedTokenizer, top_ks: List[int],
                 max_query_tokens: int, has_query_noise: bool, use_rewrite: bool = True,
                 max_scenarios: int = 2000, max_num_facts_per_scenario: int = 1):
        super().__init__(tokenizer, top_ks, max_query_tokens, has_query_noise)
        self.max_num_facts_per_scenario = max_num_facts_per_scenario
        self.use_rewrite = use_rewrite
        data_dir = download_zip(_url)
        data_file = os.path.join(data_dir, _test_file)
        with codecs.open(data_file, 'r', 'utf-8') as fd:
            all_data = json.load(fd)
            rnd = np.random.RandomState(8011)
            rnd.shuffle(all_data)
            self.data = all_data[:max_scenarios]

    def get_facts_to_be_stored(self) -> List[str]:
        facts = []
        for entry in self.data:
            entry_facts = entry.get('facts')
            if entry_facts:
                facts.extend(entry_facts[:self.max_num_facts_per_scenario])
        return facts

    def get_scenarios(self) -> List[QAScenario]:
        scenarios = []
        for entry in self.data:
            entry_facts = entry.get('facts')
            if entry_facts:
                description = entry.get('description', '')
                q = entry.get('question')
                if q:
                    context = [description]
                    s_facts = entry_facts[:self.max_num_facts_per_scenario]
                    scenario = QAScenario(context=context, supportingFacts=s_facts, question=q)
                    scenarios.append(scenario)
        return scenarios



from abc import ABC, abstractmethod

import torch
from torch import nn


class BaseQueryPassageProbModel(ABC):
    """
    Abstract base class for query-passage probability models.
    """

    @abstractmethod
    def forward(self, query_input_ids: torch.Tensor, query_attention_mask: torch.Tensor,
                query_token_lengths: torch.Tensor,
                passage_input_ids: torch.Tensor, passage_attention_mask: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def get_lm_parameters(self):
        pass

    @abstractmethod
    def get_added_parameters(self):
        pass

    @abstractmethod
    def train(self, mode: bool = True):
        pass



from abc import ABC, abstractmethod
from pathlib import Path

from goodai.ltm.mem.default import DefaultTextMemory


class MemoryPersistence(ABC):
    """
    An interface that defines various strategies for persisting the memory to disk.
    This abstract base class ensures that any derived class implements both the
    save and load methods for memory persistence.
    """

    @abstractmethod
    def save(self, memory: DefaultTextMemory, directory: Path) -> None:
        """
        Save the provided memory to the specified directory.

        :param memory: The memory instance to be saved.
        :type memory: DefaultTextMemory
        :param directory: The directory where the memory should be saved.
        :type directory: Path
        """
        pass

    @abstractmethod
    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        """
        Load the memory from the specified directory.

        :param directory: The directory from which the memory should be loaded.
        :type directory: Path
        :param kwargs: Additional keyword arguments passed to AutoTextMemory.create.
        :return: The loaded memory instance.
        :rtype: DefaultTextMemory
        """
        pass

    @abstractmethod
    def exists(self, directory: Path) -> bool:
        """
        Check whether the memory exists in the specified directory.

        :param directory: The directory to check.
        :type directory: Path
        :return: True if the memory exists, False otherwise.
        :rtype: bool
        """
        pass



import codecs
import json
import os
from typing import List, Optional

import datasets
import numpy as np
from transformers import PreTrainedTokenizer

from goodai.helpers.file_helper import download_zip
from goodai.ltm.data.query_passage.auto_data_source import AutoQueryPassageDataSource
from goodai.ltm.eval.mem import BaseMemEvaluator, QAScenario


class QPDSMemEvaluator(BaseMemEvaluator):
    # For testing query-passage data sources
    def __init__(self, ds_name: str, tokenizer: PreTrainedTokenizer, top_ks: List[int],
                 max_query_tokens: int, has_query_noise: bool, use_rewrite: bool = True,
                 max_scenarios: int = 2000):
        super().__init__(tokenizer, top_ks, max_query_tokens, has_query_noise, add_names_to_context=False)
        self.use_rewrite = use_rewrite
        rnd = np.random.RandomState(7089)
        _, ds = AutoQueryPassageDataSource.create(ds_name, rnd, tokenizer, max_query_tokens=max_query_tokens)
        examples = ds.sample_items(max_scenarios, approx_positive_fraction=1.0)
        entries = []
        for ex in examples:
            query = tokenizer.decode(ex.queryIds, skip_special_tokens=True)
            passage = tokenizer.decode(ex.passageIds, skip_special_tokens=True)
            entries.append((query, passage,))
        self.entries = entries

    def get_facts_to_be_stored(self) -> List[str]:
        return [p for _, p in self.entries]

    def get_scenarios(self) -> List[QAScenario]:
        scenarios = []
        for q, p in self.entries:
            scenario = QAScenario(context=[], supportingFacts=[p], question=q)
            scenarios.append(scenario)
        return scenarios



import gc
import itertools
import math
from typing import List, Tuple, Optional

import numpy as np
import torch
from torch import nn

from goodai.helpers.sched_opt import ScheduledOptimizer
from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource
from goodai.ltm.data.query_passage.auto_data_source import AutoQueryPassageDataSource
from torch.utils.data import DataLoader, Dataset
from transformers import PreTrainedTokenizer
from timeit import default_timer as timer

from goodai.ltm.data.query_passage.dataset import QueryPassageDataset
from goodai.ltm.embeddings.trainable import TrainableEmbeddingModel
from goodai.modules.loss import EmbCrossProbLossModel


class EmbModelTrainer:
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                 num_epochs: int, switch_ds_every: int, num_ds_examples: int, batch_size: int,
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 track_validation: bool, lm_lr: float, extras_lr: float, device: torch.device,
                 num_warmup_steps: int = 0, weight_decay: float = 1e-3):
        super().__init__()
        self.weight_decay = weight_decay
        self.num_warmup_steps = num_warmup_steps
        self.extras_lr = extras_lr
        self.lm_lr = lm_lr
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.track_validation = track_validation
        self.batch_size = batch_size
        self.num_ds_examples = num_ds_examples
        self.switch_ds_every = switch_ds_every
        self.num_epochs = num_epochs
        self.tokenizer = tokenizer
        self.random = random
        self.device = device
        self.train_data_sources: List[Tuple[BaseQueryPassageDataSource, float]] = []
        self.valid_data_sources: List[Tuple[BaseQueryPassageDataSource, float]] = []

    def add_data_source(self, ds_name: str, weight: float):
        train_ds, valid_ds = AutoQueryPassageDataSource.create(ds_name, self.random, self.tokenizer,
                                                               max_query_tokens=self.max_query_tokens,
                                                               min_passage_tokens=self.min_passage_tokens,
                                                               max_passage_tokens=self.max_passage_tokens)
        self.train_data_sources.append((train_ds, weight,))
        self.valid_data_sources.append((valid_ds, weight,))

    def train_dataset(self, dataset: Dataset, model: TrainableEmbeddingModel, loss_model: nn.Module,
                      s_opt: Optional[ScheduledOptimizer],
                      validation=False) -> Tuple[float, float]:
        training = not validation
        model.train(mode=training)
        loss_sum = 0
        bp_time = 0
        pass_count = 0
        data_loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)
        for i in range(self.switch_ds_every):
            for q_tuple, p_tuple, _ in data_loader:
                qid, qtl, qam = q_tuple
                pid, pam = p_tuple
                a_embeddings = model(input_ids=qid, attention_mask=qam, token_lengths=qtl, is_retrieve=True)
                b_embeddings = model(input_ids=pid, attention_mask=pam, token_lengths=None, is_retrieve=False)
                loss = loss_model(a_embeddings, b_embeddings)
                loss_sum += loss.item()
                if training:
                    s_opt.zero_grad()
                    time1 = timer()
                    gc.collect()
                    loss.backward()
                    time2 = timer()
                    bp_time += time2 - time1
                    s_opt.step()
                    s_opt.zero_grad(set_to_none=True)
                del a_embeddings
                del b_embeddings
                del loss
                gc.collect()
                pass_count += 1
        return loss_sum / pass_count, bp_time,

    def train(self, model: TrainableEmbeddingModel):
        print('------------------------------------')
        print(f'ML learning rate: {self.lm_lr}')
        print(f'Extras learning rate: {self.extras_lr}')
        print(f'Batch size: {self.batch_size}')
        print(f'Num epochs per dataset: {self.switch_ds_every}')
        print(f'Num epochs: {self.num_epochs}')
        print(f'Max query tokens: {self.max_query_tokens}')
        print(f'Min passage tokens: {self.min_passage_tokens}')
        print(f'Max passage tokens: {self.max_passage_tokens}')
        print('------------------------------------')
        if self.num_epochs > 0:
            print('Training...')
            loss_model = EmbCrossProbLossModel().to(self.device)
            lm_parameters = model.get_lm_parameters()
            extra_parameters = model.get_added_parameters()
            all_extra_parameters = itertools.chain(extra_parameters, loss_model.parameters())
            num_training_steps = self.num_epochs * math.ceil(self.num_ds_examples / self.batch_size)
            parameters = [
                {'params': lm_parameters, 'lr': self.lm_lr},
                {'params': all_extra_parameters, 'lr': self.extras_lr},
            ]
            s_opt = ScheduledOptimizer(parameters, lr=self.lm_lr,
                                       num_training_steps=num_training_steps,
                                       num_warmup_steps=self.num_warmup_steps,
                                       weight_decay=self.weight_decay)
            sum_bp_time = 0
            time1 = timer()
            for epoch in range(0, self.num_epochs, self.switch_ds_every):
                dataset = self.create_dataset(training=True)
                train_loss, bp_time = self.train_dataset(dataset, model, loss_model, s_opt, validation=False)
                sum_bp_time += bp_time
                print(f'-- Epoch {epoch}: Train Loss={train_loss:.4g}')
                last_lrs = s_opt.get_last_lrs()
                print(f'Last LM LR: {np.round(last_lrs, 6)}')
                if self.track_validation and epoch % 20 == 0:
                    valid_ds = self.create_dataset(training=False)
                    valid_loss, _ = self.train_dataset(valid_ds, model, loss_model, None,
                                                       validation=True)
                    print(f'== Epoch {epoch}: Valid Loss={valid_loss:.4g}')
            time2 = timer()
            elapsed = time2 - time1
            print(f'Trained {self.num_epochs} epochs in {elapsed:.1f} seconds.')
            bp_time_pct = sum_bp_time * 100.0 / elapsed
            print(f'Backprop time: {sum_bp_time:.1f} seconds ({bp_time_pct:.1f}%)')
            eph = self.num_epochs / (elapsed / (60 * 60))
            print(f'Training speed: {eph:.1f} epochs per hour.')
            torch.cuda.empty_cache()
            gc.collect()

    def create_dataset(self, training: bool):
        data_sources = self.train_data_sources if training else self.valid_data_sources
        actual_ds_examples = self.num_ds_examples
        return QueryPassageDataset(data_sources, self.tokenizer, actual_ds_examples,
                                   device=self.device, approx_positive_fraction=1.0)



from abc import ABC, abstractmethod
from typing import List
from goodai.ltm.data.query_passage.example import QueryPassageExample


class BaseQueryPassageDataSource(ABC):
    """
    Abstract interface for data sources for training and evaluation of query-passage matching models.
    """

    def __init__(self):
        pass

    @abstractmethod
    def sample_items(self, count: int, approx_positive_fraction: float = 0.5) -> List[QueryPassageExample]:
        pass



import os
from typing import List

import numpy as np

from goodai.helpers.json_helper import load_json

_instance_attr = '__instance'
_names_dir = os.path.join(os.path.dirname(__file__))
_names_path = os.path.join(_names_dir, 'wikidata-names.json')


class NameSource:
    def __init__(self):
        def _filter(names: List[str]):
            return [n for n in names if len(n) <= 10]

        names_obj = load_json(_names_path)
        self.male_names = np.array(_filter(names_obj['MALE']))
        self.female_names = np.array(_filter(names_obj['FEMALE']))
        self.family_names = np.array(_filter(names_obj['FAMILY']))

    @classmethod
    def get_instance(cls):
        try:
            return getattr(cls, _instance_attr)
        except AttributeError:
            instance = cls()
            setattr(cls, _instance_attr, instance)
            return instance

    def get_male_names(self) -> List[str]:
        return list(self.male_names)

    def get_female_names(self) -> List[str]:
        return list(self.female_names)

    def get_family_names(self) -> List[str]:
        return list(self.family_names)

    def sample_male_names(self, random: np.random.RandomState, count: int):
        return random.choice(self.male_names, size=count, replace=False)

    def sample_female_names(self, random: np.random.RandomState, count: int):
        return random.choice(self.female_names, size=count, replace=False)

    def sample_family_names(self, random: np.random.RandomState, count: int):
        return random.choice(self.family_names, size=count, replace=False)

    def sample_first_names(self, random: np.random.RandomState, count: int):
        male_names = self.sample_male_names(random, count)
        female_names = self.sample_female_names(random, count)
        is_male = random.choice([True, False], size=count, replace=True)
        return [m if im else f for m, f, im in zip(male_names, female_names, is_male)]



from typing import List

from transformers import PreTrainedTokenizer


def levenshtein_distance(s_pred: List[str], t_exp: List[str]):
    m = len(s_pred)
    n = len(t_exp)
    d = [[0] * (n + 1) for _ in range(m + 1)]
    for i in range(m + 1):
        d[i][0] = i
    for j in range(n + 1):
        d[0][j] = j
    for j in range(1, n + 1):
        for i in range(1, m + 1):
            if s_pred[i - 1] == t_exp[j - 1]:
                sub_cost = 0
            else:
                sub_cost = 1
            d[i][j] = min([
                d[i - 1][j] + 1,  # Deletion
                d[i][j - 1] + 1,  # Insertion
                d[i - 1][j - 1] + sub_cost,  # Substitution
            ])
    return d[m][n]


def get_correctness_score(tokenizer: PreTrainedTokenizer, predicted_answer: str, expected_answer: str):
    predicted_raw_tokens = tokenizer.tokenize(' ' + predicted_answer, add_special_tokens=False)
    return _get_correctness_score_for_tokens_ea(tokenizer, predicted_raw_tokens, expected_answer)


def _get_correctness_score_for_tokens_ea(tokenizer: PreTrainedTokenizer, predicted_raw_tokens: List[str],
                                         expected_answer: str):
    expected_raw_tokens_2 = tokenizer.tokenize(' ' + expected_answer, add_special_tokens=False)
    correctness_2 = _get_correctness_score_for_tokens(predicted_raw_tokens, expected_raw_tokens_2)
    return correctness_2


def _get_correctness_score_for_tokens(predicted_raw_tokens: List[str], expected_raw_tokens: List[str]):
    predicted_tokens = _norm_tokens(predicted_raw_tokens)
    expected_tokens = _norm_tokens(expected_raw_tokens)
    if len(expected_tokens) == 0:
        return 0.0
    else:
        s_distance = _subseq_distance(predicted_tokens, expected_tokens)
        max_distance = len(expected_tokens)
        return 100.0 * (1 - s_distance / max_distance)


def _subseq_distance(s_pred: List[str], t_exp: List[str]):
    min_s_len = len(t_exp)
    i_max = len(s_pred) - min_s_len
    if i_max <= 0:
        pred_subseqs = [s_pred]
    else:
        pred_subseqs = [s_pred[i:i + min_s_len] for i in range(i_max + 1)]
    distances = [levenshtein_distance(s, t_exp) for s in pred_subseqs]
    return min(distances)


def _norm_token(token: str):
    if token.startswith('Ġ'):
        token = token[1:]
    token = token.lower()
    if token in [',', '!', ';', '..', '...', '!!', '!!!']:
        token = '.'
    return token


def _norm_tokens(raw_tokens: List[str]):
    tokens = [_norm_token(t) for t in raw_tokens]
    tokens = [t for t in tokens if t != '']
    if tokens[-1:] == ['.']:
        tokens = tokens[:-1]
    return tokens



import bisect
import logging
import time
from dataclasses import dataclass, field
from typing import List, Dict, Optional
from typing import Tuple, Any, Set

import numpy as np
from transformers import PreTrainedTokenizer

from goodai.ltm.mem.chunk import Chunk, TextKeyType
from goodai.ltm.mem.config import ChunkExpansionConfig


@dataclass
class PassageInfo:
    fromIndex: int
    toIndex: int
    tokenIds: List[int]


@dataclass
class ChunkExpansionOptions:
    minSideTokens: int
    maxSideTokens: int
    leftStopAfterTokenIds: List[List[int]]
    rightStopAtTokenIds: List[List[int]]

    @classmethod
    def default(cls, max_side_tokens: int, punctuation_ids: Set[int]):
        token_ids = [[tid] for tid in punctuation_ids]
        return cls(minSideTokens=0, maxSideTokens=max_side_tokens,
                   leftStopAfterTokenIds=token_ids,
                   rightStopAtTokenIds=token_ids)

    @classmethod
    def from_config(cls, tokenizer: PreTrainedTokenizer, config: ChunkExpansionConfig):
        celt = config.limit_type
        limit_token_ids = celt.get_token_ids(tokenizer)
        return cls(minSideTokens=config.min_extra_side_tokens,
                   maxSideTokens=config.max_extra_side_tokens,
                   leftStopAfterTokenIds=limit_token_ids,
                   rightStopAtTokenIds=limit_token_ids)


@dataclass
class ChunkQueue:
    capacity: int
    chunk_capacity: int
    chunk_index_at_overlap: int
    first_token_seq_id: int = 0
    chunks: List[Chunk] = field(default_factory=list)
    current_chunk_id: int = 0
    current_text_key: int = 0
    token_ids: List[int] = field(default_factory=list)
    separator_seq_ids: List[int] = field(default_factory=list)
    sequence_map: Dict[TextKeyType, Tuple[int, int]] = field(default_factory=dict)
    chunk_map: Dict[int, Chunk] = field(default_factory=dict, init=False)

    def __post_init__(self):
        if self.capacity <= 2:
            raise ValueError('Queue capacity cannot be 2 or less')
        if self.chunk_capacity < 1:
            raise ValueError('Chunk capacity cannot be zero or less')
        if self.chunk_index_at_overlap < self.chunk_capacity // 2:
            raise ValueError('Chunk overlap cannot be more than 50%')
        self.chunk_map = {c.chunk_id: c for c in self.chunks}

    def _removed_chunk_cleanup(self, chunk: Chunk):
        self.chunk_map.pop(chunk.chunk_id)
        if len(self.chunks) == 0:
            assert len(self.chunk_map) == 0
            self.flush()
        else:
            new_first_chunk = self.chunks[0]
            new_first_token_seq_id = new_first_chunk.from_token_seq_id
            num_removed = new_first_token_seq_id - self.first_token_seq_id
            if num_removed != 0:
                self.token_ids = self.token_ids[num_removed:]
                self.first_token_seq_id = new_first_token_seq_id
                sii = bisect.bisect_left(self.separator_seq_ids, new_first_token_seq_id)
                self.separator_seq_ids = self.separator_seq_ids[sii:]
                for text_key in chunk.associated_keys:
                    text_bounds = self.sequence_map.get(text_key)
                    if text_bounds is not None:
                        _, text_to = text_bounds
                        if text_to <= new_first_token_seq_id:
                            del self.sequence_map[text_key]

    def _pop_chunk(self):
        chunk = self.chunks.pop(0)
        if chunk is not None:
            self._removed_chunk_cleanup(chunk)
        return chunk

    def get_all_chunks(self) -> List[Chunk]:
        return list(self.chunks)

    def check_overflow(self) -> List[Chunk]:
        removed_chunks = []
        while len(self.chunks) > self.capacity:
            removed_chunk = self._pop_chunk()
            if removed_chunk is not None:
                removed_chunks.append(removed_chunk)
        return removed_chunks

    def add_chunk(self, metadata: Optional[Any], importance: Optional[float], timestamp: float,
                  starts_section: bool = False, section_seq_id: Optional[int] = None) -> Chunk:
        chunk_id = self.current_chunk_id
        last_chunk = self.chunks[-1] if len(self.chunks) >= 1 else None
        if last_chunk is None:
            from_token_seq_id = self.first_token_seq_id
        elif starts_section:
            assert section_seq_id is not None
            from_token_seq_id = section_seq_id
        else:
            from_token_seq_id = last_chunk.from_token_seq_id + self.chunk_index_at_overlap
        chunk = Chunk(chunk_id, self.chunk_capacity, from_token_seq_id, metadata, importance, timestamp)
        self.current_chunk_id = chunk_id + 1
        self.chunks.append(chunk)
        self.chunk_map[chunk.chunk_id] = chunk
        return chunk

    def get_chunk(self, chunk_id: int) -> Chunk:
        return self.chunk_map.get(chunk_id)

    def extend_chunk(self, chunk: Chunk, token_ids: List[int], new_tokens: bool):
        if new_tokens:
            self.token_ids.extend(token_ids)
        chunk.extend_by(len(token_ids))

    def get_first_token_sequence_id(self) -> int:
        return self.first_token_seq_id

    def get_next_token_sequence_id(self) -> int:
        return self.first_token_seq_id + len(self.token_ids)

    def _new_text_key(self) -> TextKeyType:
        self.current_text_key += 1
        return self.current_text_key

    def _pad_last_chunk(self, pad_token_id: int):
        last_chunk = self.chunks[-1] if len(self.chunks) >= 1 else None
        if last_chunk is not None:
            room = last_chunk.get_room()
            pad_seq = [pad_token_id] * room
            self.add_sequence(pad_seq, metadata=None, importance=None, _no_new_chunks=True)
            return room
        else:
            return 0

    def _update_sequence_map(self, text_key: TextKeyType, new_to_seq_id: int, shift_offset: int):
        # TODO not super efficient
        old_bounds = self.sequence_map.get(text_key)
        if old_bounds is None:
            return
        from_seq_id, old_to_seq_id = old_bounds

        def _shift(b_from_id: int, b_to_id: int):
            if b_from_id >= old_to_seq_id:
                return b_from_id + shift_offset, b_to_id + shift_offset,
            else:
                return b_from_id, b_to_id,

        self.sequence_map = {k: _shift(*b) for k, b in self.sequence_map.items()}
        self.sequence_map[text_key] = (from_seq_id, new_to_seq_id,)

    def _update_separator_ids(self, remove_from_id: int, remove_to_id: int, shift_offset: int):
        old_ids = self.separator_seq_ids
        if len(old_ids) > 0:
            idx1 = bisect.bisect_right(old_ids, remove_from_id)
            idx2 = bisect.bisect_left(old_ids, remove_to_id)
            self.separator_seq_ids = old_ids[:idx1] + \
                                     (np.array(old_ids[idx2:], dtype=int) + shift_offset).tolist()

    def _split_separators(self, seq_from_id: int, seq_to_id: int) -> Tuple[List[int], List[int]]:
        old_ids = self.separator_seq_ids
        if len(old_ids) > 0:
            idx1 = bisect.bisect_right(old_ids, seq_from_id)
            idx2 = bisect.bisect_left(old_ids, seq_to_id)
            return old_ids[:idx1], old_ids[idx2:],
        else:
            return [], [],

    def _is_start_section(self):
        sep_ids = self.separator_seq_ids
        if len(sep_ids) > 0:
            last_sep_id = sep_ids[-1]
            if len(self.token_ids) + self.first_token_seq_id == last_sep_id:
                return True
        return False

    def add_separator(self):
        # self._pad_last_chunk(pad_token_id)
        # separator_seq_ids always assumed to be ordered
        self.separator_seq_ids.append(self.first_token_seq_id + len(self.token_ids))

    def add_sequence(self, new_token_ids: List[int], metadata: Optional[Any],
                     importance: Optional[float] = None, timestamp: Optional[float] = None,
                     _no_new_chunks: bool = False,
                     _no_text_key: bool = False,
                     _text_key: TextKeyType = None) -> Tuple[List[Chunk], TextKeyType]:
        """
        Adds tokens to the chunk queue.
        :param new_token_ids: The sequence of token IDs to add
        :param metadata: A metadata object
        :param importance: An optional importance value for the tokens
        :param timestamp: The timestamp of the stored tokens
        :param _no_new_chunks: If true, attempt should be made to add all tokens without adding new chunks
        :param _no_text_key: Whether associating text with a key should be prevented.
        :param _text_key: A text key to be used instead of a generated one.
        :return: Any chunks removed due to overflow.
        """
        if len(new_token_ids) == 0:
            return [], _text_key,
        if timestamp is None:
            timestamp = time.time()
        text_key = None if _no_text_key else (_text_key if _text_key is not None else self._new_text_key())
        first_id = self.first_token_seq_id
        prev_token_seq_id = len(self.token_ids) + first_id
        starts_section = self._is_start_section()
        self.token_ids.extend(new_token_ids)
        next_token_seq_id = len(self.token_ids) + first_id
        start_num_chunks = len(self.chunks)
        first_c_idx = max(0, start_num_chunks if starts_section else start_num_chunks - 2)
        key_registered = False
        new_chunk_starts_section = starts_section
        for c_idx in range(first_c_idx, start_num_chunks + len(new_token_ids) + 1):
            if _no_new_chunks:
                if c_idx >= len(self.chunks):
                    raise RuntimeError('No new chunks allowed, but at least one more is needed to complete operation')
            else:
                while c_idx >= len(self.chunks):
                    self.add_chunk(metadata, importance, timestamp=timestamp, starts_section=new_chunk_starts_section,
                                   section_seq_id=prev_token_seq_id)
                    new_chunk_starts_section = False
            chunk = self.chunks[c_idx]
            if chunk.to_token_seq_id < next_token_seq_id:
                room = chunk.get_room()
                if room > 0:
                    if not _no_text_key:
                        chunk.add_key(text_key)
                        key_registered = True
                    chunk.extend_by(min(next_token_seq_id - chunk.to_token_seq_id, room))
                    much_room = room > self.chunk_capacity // 2
                    if metadata and (chunk.metadata is None or much_room):
                        chunk.metadata = metadata
                    if importance is not None and (chunk.importance is None or much_room):
                        chunk.importance = importance
                    if room == self.chunk_capacity:
                        chunk.timestamp = timestamp
            if (_no_new_chunks or len(chunk) <= self.chunk_index_at_overlap) and \
                    chunk.to_token_seq_id >= next_token_seq_id:
                break
        if key_registered:
            self.sequence_map[text_key] = (prev_token_seq_id, next_token_seq_id,)
        return self.check_overflow(), text_key,

    def _split_left(self, head_chunks: List[Chunk], tail_id_from: int):
        first_id = self.first_token_seq_id
        if len(head_chunks) > 0:
            last_head_chunk = head_chunks[-1]
            last_head_chunk_to = last_head_chunk.to_token_seq_id
            left_extras_from = last_head_chunk_to - first_id
            left_extras_to = tail_id_from - first_id
            left_extras_token_ids = self.token_ids[left_extras_from:left_extras_to]
            head_token_ids = self.token_ids[:left_extras_from]
            chunk_params = dict(
                timestamp=last_head_chunk.timestamp,
                importance=last_head_chunk.importance,
                metadata=last_head_chunk.metadata,
            )
        else:
            left_extras_token_ids = []
            head_token_ids = self.token_ids[:tail_id_from - first_id]
            chunk_params = dict(metadata=None)
        return head_token_ids, left_extras_token_ids, chunk_params,

    def _split_right(self, shifted_chunks: List[Chunk], head_id_to: int):
        first_id = self.first_token_seq_id
        if len(shifted_chunks) > 0:
            first_shifted_chunk = shifted_chunks[0]
            first_shifted_chunk_from = first_shifted_chunk.from_token_seq_id
            right_extras_from = head_id_to - first_id
            right_extras_to = first_shifted_chunk_from - first_id
            right_extras_token_ids = self.token_ids[right_extras_from:right_extras_to]
            shifted_token_ids = self.token_ids[right_extras_to:]
            chunk_params = dict(
                timestamp=first_shifted_chunk.timestamp,
                importance=first_shifted_chunk.importance,
                metadata=first_shifted_chunk.metadata,
            )
        else:
            right_extras_token_ids = []
            shifted_token_ids = self.token_ids[head_id_to:]
            chunk_params = dict(metadata=None)
        return shifted_token_ids, right_extras_token_ids, chunk_params,

    def _resolve_discarded_chunks(self, discarded_chunks: List[Chunk], text_key: TextKeyType,
                                  seq_id_from: int, old_seq_id_to: int, new_seq_id_to: int,
                                  shift_offset: int):
        # Remove chunks from map, etc.
        for chunk in discarded_chunks:
            self._removed_chunk_cleanup(chunk)
        # Update the bounds of the current sequence, and shift the ones to the right
        self._update_sequence_map(text_key, new_seq_id_to, shift_offset)
        # Remove and shift separator seq IDs
        self._update_separator_ids(seq_id_from, old_seq_id_to, shift_offset)
        return discarded_chunks

    def replace_sequence(self, text_key: TextKeyType, new_token_ids: List[int],
                         metadata: Optional[dict] = None, importance: Optional[float] = None,
                         timestamp: Optional[float] = None) -> Tuple[List[Chunk], TextKeyType]:
        old_bounds = self.sequence_map.get(text_key)
        if old_bounds is None:
            logging.warning(f'Subsequence with key {text_key} not found.')
            return [], text_key,
        seq_id_from, old_seq_id_to = old_bounds
        new_seq_id_to = seq_id_from + len(new_token_ids)
        discarded_chunks = []
        shifted_chunks = []
        head_chunks = []
        for chunk in self.chunks:
            if chunk.to_token_seq_id > seq_id_from:
                if chunk.from_token_seq_id >= old_seq_id_to:
                    shifted_chunks.append(chunk)
                else:
                    discarded_chunks.append(chunk)
            else:
                head_chunks.append(chunk)
        # Find out which tokens will be left out prior to first shifted chunk, plus overlap
        shifted_token_ids, right_extras_token_ids, right_params = self._split_right(shifted_chunks, old_seq_id_to)
        # Find out which tokens will be left out after the last head chunk, before the new sequence
        head_token_ids, left_extras_token_ids, left_params = self._split_left(head_chunks, seq_id_from)
        # Split separator IDs
        head_sep_ids, shifted_sep_ids = self._split_separators(seq_id_from, old_seq_id_to)
        # Truncate chunks, token IDs and separators
        self.token_ids = head_token_ids
        self.chunks = head_chunks
        self.separator_seq_ids = head_sep_ids
        # Add new token sequence to truncated queues
        self.add_sequence(left_extras_token_ids, _no_text_key=True, **left_params)
        self.add_sequence(new_token_ids, metadata=metadata, importance=importance,
                          timestamp=timestamp)
        self.add_sequence(right_extras_token_ids, _no_text_key=True, **right_params)
        # Calculate the shift offset
        shift_offset = new_seq_id_to - old_seq_id_to
        # Shift the sequence IDs of the chunks that need to be shifted
        # No padding is added, so shifted chunks should contain the same tokens as before
        for chunk in shifted_chunks:
            chunk.shift(shift_offset)
        # Extend the token IDs queue and the chunk queue
        self.token_ids.extend(shifted_token_ids)
        self.chunks.extend(shifted_chunks)
        self.separator_seq_ids.extend(shifted_sep_ids)
        # This call updates separator seq IDs according to the shift offset
        dc1 = self._resolve_discarded_chunks(discarded_chunks, text_key,
                                             seq_id_from, old_seq_id_to, new_seq_id_to,
                                             shift_offset)
        dc2 = self.check_overflow()
        return dc1 + dc2, text_key,

    def get_sequence_token_ids(self, text_key: TextKeyType) -> Optional[List[int]]:
        seq_bounds = self.sequence_map.get(text_key)
        if seq_bounds is None:
            logging.warning(f'Subsequence with key {text_key} not found.')
            return None
        seq_id_from, seq_id_to = seq_bounds
        first_id = self.first_token_seq_id
        index_from, index_to = seq_id_from - first_id, seq_id_to - first_id,
        return self.token_ids[index_from:index_to]

    def get_chunks_for_indexing(self) -> Tuple[List[Chunk], List[List[int]]]:
        # TODO not super efficient
        token_id_matrix = []
        picked_buckets = []
        for i, chunk in enumerate(self.chunks):
            if not chunk.is_indexed():
                token_ids = self.get_chunk_token_ids(chunk)
                if len(token_ids) > 0:
                    picked_buckets.append(chunk)
                    token_id_matrix.append(token_ids)
        return picked_buckets, token_id_matrix,

    def get_chunk_token_ids(self, chunk: Chunk):
        first_index = self.first_token_seq_id
        from_index = chunk.from_token_seq_id - first_index
        to_index = chunk.to_token_seq_id - first_index
        return self.token_ids[from_index:to_index]

    def _get_section_bounds(self, from_index: int, to_index: int) -> Tuple[int, int]:
        ssi = bisect.bisect_right(self.separator_seq_ids, from_index)
        ss_seq_id = self.separator_seq_ids[ssi - 1] if ssi >= 1 else self.first_token_seq_id
        esi = bisect.bisect_left(self.separator_seq_ids, to_index)
        es_seq_id = self.separator_seq_ids[esi] if esi < len(self.separator_seq_ids) \
            else self.first_token_seq_id + len(self.token_ids)
        return ss_seq_id, es_seq_id,

    def _get_prev_token_ids(self, chunk: Chunk, section_from_seq_id: int) -> Optional[List[int]]:
        to_index = chunk.from_token_seq_id - self.first_token_seq_id
        from_index = to_index - self.chunk_capacity
        from_index = max(section_from_seq_id - self.first_token_seq_id, from_index)
        return self.token_ids[from_index:to_index]

    def _get_next_token_ids(self, chunk: Chunk, section_to_seq_id: int) -> Optional[List[int]]:
        from_index = chunk.to_token_seq_id - self.first_token_seq_id
        to_index = min(section_to_seq_id - self.first_token_seq_id, from_index + self.chunk_capacity)
        return self.token_ids[from_index:to_index]

    def retrieve_complete_sequences(self, chunk_ids: List[int], options: ChunkExpansionOptions) -> List[List[int]]:
        sequences = []
        for chunk_id in chunk_ids:
            chunk = self.chunk_map.get(chunk_id)
            if chunk is not None:
                passage = self.get_complete_passage(chunk, options)
                sequence = passage.tokenIds
            else:
                sequence = []
            sequences.append(sequence)
        return sequences

    def get_complete_passage(self, chunk: Chunk, options: ChunkExpansionOptions) -> PassageInfo:
        s_from, s_to = self._get_section_bounds(chunk.from_token_seq_id, chunk.to_token_seq_id)
        min_st = options.minSideTokens
        min_p_from, min_p_to = chunk.from_token_seq_id - min_st, chunk.to_token_seq_id + min_st,
        min_p_from, min_p_to = max(min_p_from, s_from), min(min_p_to, s_to)
        max_st = options.maxSideTokens
        max_p_from, max_p_to = chunk.from_token_seq_id - max_st, chunk.to_token_seq_id + max_st,
        max_p_from, max_p_to = max(max_p_from, s_from), min(max_p_to, s_to)
        central_seq = self.get_subsequence(min_p_from, min_p_to)
        prev_token_ids = self.get_subsequence(max_p_from, min_p_from)
        prev_token_ids = self._from_last_match(prev_token_ids, options.leftStopAfterTokenIds)
        next_token_ids = self.get_subsequence(min_p_to, max_p_to)
        next_token_ids = self._to_first_match(next_token_ids, options.rightStopAtTokenIds)
        p_from = min_p_from - len(prev_token_ids)
        p_to = min_p_to + len(next_token_ids)
        expanded_seq = prev_token_ids + central_seq + next_token_ids
        return PassageInfo(p_from, p_to, expanded_seq)

    def get_subsequence(self, from_seq_id: int, to_seq_id: int):
        first = self.first_token_seq_id
        from_idx = from_seq_id - first
        to_idx = to_seq_id - first
        return self.token_ids[from_idx:to_idx]

    @staticmethod
    def _from_last_match(token_ids: List[int], sub_seqs: List[List[int]]):
        nt = len(token_ids)
        match_indexes = []
        for sub_seq in sub_seqs:
            ls = len(sub_seq)
            if ls > 0:
                for i in range(nt - ls, -1, -1):
                    if token_ids[i:i + ls] == sub_seq:
                        match_indexes.append((i, ls,))
                        break
        if len(match_indexes) == 0:
            return token_ids
        max_idx, max_len = max(match_indexes, key=lambda _t: _t[0])
        return token_ids[max_idx + max_len:]

    @staticmethod
    def _to_first_match(token_ids: List[int], sub_seqs: List[List[int]]):
        nt = len(token_ids)
        match_indexes = []
        for sub_seq in sub_seqs:
            ls = len(sub_seq)
            if ls > 0:
                for i in range(0, nt - ls):
                    if token_ids[i:i + ls] == sub_seq:
                        match_indexes.append((i, ls,))
                        break
        if len(match_indexes) == 0:
            return token_ids
        min_idx, min_len = min(match_indexes, key=lambda _t: _t[0])
        return token_ids[:min_idx + min_len]

    def get_queue_size(self):
        return len(self.chunks)

    def get_chunk_sequences(self) -> List[List[int]]:
        return [self.get_chunk_token_ids(chunk) for chunk in self.chunks]

    def get_capacity(self) -> int:
        return self.capacity

    def get_latest_token_ids(self, max_num_tokens: Optional[int]):
        total = 0
        result = []
        len_queue = len(self.chunks)
        for i in range(len_queue - 1, -1, -2):
            chunk = self.chunks[i]
            token_ids = self.get_chunk_token_ids(chunk)
            new_total = total + len(token_ids)
            if max_num_tokens is not None and new_total >= max_num_tokens:
                remain_count = max_num_tokens - total
                if remain_count > 0:
                    result[:0] = token_ids[-remain_count:]
                break
            else:
                result[:0] = token_ids
            total = new_total
        return result

    def retrieve_chunk_sequences(self, chunk_ids: List[int]):
        sequences = []
        for chunk_id in chunk_ids:
            chunk = self.chunk_map.get(chunk_id)
            if chunk is not None:
                sequence = self.get_chunk_token_ids(chunk)
                sequences.append(sequence)
        return sequences

    def retrieve_chunk_sequences_given_chunks(self, chunks: List[Chunk]):
        sequences = []
        for chunk in chunks:
            sequence = self.get_chunk_token_ids(chunk)
            sequences.append(sequence)
        return sequences

    def flush(self):
        self.chunks = []
        self.token_ids = []
        self.separator_seq_ids = []
        self.first_token_seq_id = 0
        self.current_chunk_id = 0
        self.chunk_map.clear()
        self.sequence_map.clear()



import pytest

from goodai.ltm.mem.chunk_queue import ChunkQueue


def test_chunkqueue_initialization():
    # Test valid initialization
    cq = ChunkQueue(capacity=5, chunk_capacity=10, chunk_index_at_overlap=6)
    assert cq.get_queue_size() == 0

    # Test invalid initializations
    with pytest.raises(ValueError):
        ChunkQueue(capacity=2, chunk_capacity=10, chunk_index_at_overlap=6)

    with pytest.raises(ValueError):
        ChunkQueue(capacity=5, chunk_capacity=0, chunk_index_at_overlap=6)

    with pytest.raises(ValueError):
        ChunkQueue(capacity=5, chunk_capacity=10, chunk_index_at_overlap=4)


def test_chunkqueue_add_chunk():
    cq = ChunkQueue(capacity=5, chunk_capacity=10, chunk_index_at_overlap=6)
    chunk = cq.add_chunk(metadata={"type": "test"}, importance=0.5, timestamp=123.456)
    assert chunk.chunk_id == 0
    assert chunk.capacity == 10
    assert chunk.metadata == {"type": "test"}
    assert chunk.importance == 0.5
    assert chunk.timestamp == 123.456


def test_chunkqueue_check_overflow():
    cq = ChunkQueue(capacity=3, chunk_capacity=10, chunk_index_at_overlap=6)
    for i in range(5):
        cq.add_chunk(metadata=None, importance=None, timestamp=123.456)
    removed_chunks = cq.check_overflow()
    assert len(removed_chunks) == 2
    assert cq.get_queue_size() == 3


def test_chunkqueue_add_sequence():
    cq = ChunkQueue(capacity=3, chunk_capacity=10, chunk_index_at_overlap=6)
    removed_chunks, text_key = cq.add_sequence([1, 2, 3, 4, 5], metadata=None, importance=None)
    assert len(removed_chunks) == 0
    assert text_key == 1
    assert cq.get_next_token_sequence_id() == 5


def test_chunkqueue_get_sequence_token_ids():
    cq = ChunkQueue(capacity=3, chunk_capacity=10, chunk_index_at_overlap=6)
    _, text_key = cq.add_sequence([1, 2, 3, 4, 5], metadata=None, importance=None)
    token_ids = cq.get_sequence_token_ids(text_key)
    assert token_ids == [1, 2, 3, 4, 5]


def test_chunkqueue_flush():
    cq = ChunkQueue(capacity=3, chunk_capacity=10, chunk_index_at_overlap=6)
    cq.add_sequence([1, 2, 3, 4, 5], metadata=None, importance=None)
    cq.flush()
    assert cq.get_queue_size() == 0
    assert cq.get_next_token_sequence_id() == 0



import enum
import logging
import math
import time
from typing import List, Optional, Callable
from goodai.ltm.mem.base import BaseReranker, RetrievedMemory, BaseTextMemory, BaseImportanceModel
from goodai.text_gen.base import BaseTextGenerationModel
from goodai.text_gen.openai_tg import OpenAICompletionModel


class DecayType(enum.Enum):
    EXPONENTIAL = 0
    INVERSE = 1


class StanfordReranker(BaseReranker):
    def __init__(self, half_life: float,
                 multiplicative: bool = False,
                 decay_type: DecayType = DecayType.EXPONENTIAL,
                 use_importance: bool = True,
                 alpha_recency: float = 1.0, alpha_importance: float = 1.0,
                 alpha_relevance: float = 1.0,
                 time_fn: Callable[[], float] = None):
        """
        Constructs a StanfordReranker, per https://arxiv.org/pdf/2304.03442.pdf.
        :param half_life: When the elapsed time reaches this value, recency is 0.5,
        whereas recency is 1.0 when the elapsed time is zero.
        :param multiplicative: Whether the score should be calculated by weighted multiplication of the
        different components rather than weighted addition.
        :param decay_type: The type of decay applied to the recency estimation.
        :param alpha_recency: The weight of the recency score.
        :param alpha_importance: The weight of the importance score.
        :param alpha_relevance: The weight of the relevance score.
        :param use_importance: Whether the memory's importance metric should be used.
        :param time_fn: An optional timestamp function to use instead of time.time().
        """
        if time_fn is None:
            time_fn = time.time
        self.use_importance = use_importance
        self.multiplicative = multiplicative
        self.alpha_relevance = alpha_relevance
        self.alpha_importance = alpha_importance
        self.alpha_recency = alpha_recency
        self.decay_type = decay_type
        self.decay_coefficient = self._decay_coefficient(decay_type, half_life)
        self.warned_importance_none = False
        self.warned_bad_elapsed = False
        self.time_fn = time_fn

    @staticmethod
    def _decay_coefficient(decay_type: DecayType, half_life: float):
        if decay_type == DecayType.EXPONENTIAL:
            return -math.log(0.5) / half_life
        elif decay_type == DecayType.INVERSE:
            return half_life
        else:
            raise ValueError(f'Unrecognized decay type: {decay_type}')

    def _recency(self, elapsed: float):
        dt = self.decay_type
        if dt == DecayType.EXPONENTIAL:
            return math.exp(-(elapsed * self.decay_coefficient))
        elif dt == DecayType.INVERSE:
            dc = self.decay_coefficient
            return dc / (dc + elapsed)
        else:
            raise ValueError(f'Unrecognized decay type: {dt}')

    def _get_score(self, recency: float, importance: Optional[float], relevance: float, eps=1e-20):
        if self.multiplicative:
            if recency < 0:
                raise RuntimeError(f'Invalid recency: {recency}')
            if self.use_importance and (importance is None or importance < 0):
                raise RuntimeError(f'Invalid importance: {importance}')
            if relevance < 0:
                raise RuntimeError(f'Invalid relevance: {relevance}')
            recency_value = math.log(recency + eps)
            importance_value = math.log(importance + eps) if self.use_importance else None
            relevance_value = math.log(relevance + eps)
        else:
            recency_value = recency
            importance_value = importance
            relevance_value = relevance
        score = importance_value * self.alpha_importance if self.use_importance else 0
        return score + recency_value * self.alpha_recency + \
            relevance_value * self.alpha_relevance

    def rerank(self, r_memories: List[RetrievedMemory], mem: BaseTextMemory) -> List[RetrievedMemory]:
        if self.use_importance and not mem.has_importance_model():
            raise RuntimeError('This reranker requires a memory with an importance model')
        scored_list = []
        current_time = self.time_fn()
        for m in r_memories:
            importance = m.importance
            if importance is None and self.use_importance:
                if not self.warned_importance_none:
                    logging.warning('Importance value is None.')
                    self.warned_importance_none = True
                importance = 0
            elapsed = current_time - m.timestamp
            if elapsed < 0:
                if not self.warned_bad_elapsed:
                    logging.warning(f'Time elapsed for memory is invalid: {elapsed:.4g} seconds.')
                    self.warned_bad_elapsed = True
                elapsed = 0
            recency = self._recency(elapsed)
            score = self._get_score(recency, importance, m.relevance)
            scored_list.append((score, m))
        scored_list.sort(key=lambda _t: _t[0], reverse=True)
        return [m for _, m in scored_list]


class StanfordImportanceModel(BaseImportanceModel):
    def __init__(self, text_gen_model: BaseTextGenerationModel = None, prompt_template: str = None,
                 max_prompt_chars: int = 20000):
        if text_gen_model is None:
            text_gen_model = OpenAICompletionModel('text-davinci-003', max_tokens=2)
        if prompt_template is None:
            # Prompt from https://arxiv.org/pdf/2304.03442.pdf.
            prompt_template = "On the scale of 1 to 10, where 1 is purely mundane (e.g., brushing teeth, " \
                     "making bed) and 10 is extremely poignant (e.g., a break up, college acceptance), " \
                     "rate the likely poignancy of the following piece of memory.\r\n" \
                     "Memory: {mem_text}\r\nRating: "
        self.prompt_template = prompt_template
        self.text_gen_model = text_gen_model
        self.max_prompt_chars = max_prompt_chars

    def get_importance(self, mem_text: str, min_value=1.0, max_value=10.0) -> float:
        template_params = {'mem_text': mem_text}
        prompt = self.prompt_template.format(**template_params)
        if len(prompt) > self.max_prompt_chars:
            raise ValueError(f'Prompt is too large: {len(prompt)} characters.')
        response = self.text_gen_model.generate(prompt)
        try:
            response_number = float(response.strip())
            response_number = max(min_value, min(max_value, response_number))
            return (response_number - min_value) / (max_value - min_value)
        except ValueError:
            logging.warning(f'Response from text generation model ("{response}") could not be converted to a number.')
            return 0



from typing import List, Tuple

from sentence_transformers import CrossEncoder

from goodai.ltm.reranking.base import BaseTextMatchingModel


class SentenceTransformerTextMatchingModel(BaseTextMatchingModel):
    def __init__(self, model_name: str):
        self.model_name = model_name
        self.ce = CrossEncoder(model_name)

    def predict(self, sentences: List[Tuple[str, str]], batch_size: int = 32,
                show_progress_bar: bool = False) -> List[float]:
        results = self.ce.predict(sentences, batch_size=batch_size, show_progress_bar=show_progress_bar,
                                  convert_to_tensor=True, convert_to_numpy=False)
        return results.tolist()

    def get_info(self):
        return f'SentenceTransformer cross-encoder {self.model_name}'



import pickle
from typing import Union
from weakref import WeakValueDictionary

import torch

from goodai.helpers.file_helper import unpickle_downloaded_url
from goodai.ltm.embeddings.auto import AutoTextEmbeddingModel
from goodai.ltm.reranking.base import BaseTextMatchingModel
from goodai.ltm.reranking.default import DefaultRerankingCrossEncoder
from goodai.ltm.reranking.emb import EmbeddingBasedMatchingModel
from goodai.ltm.reranking.st_ce import SentenceTransformerTextMatchingModel

_default_dist_param = 0.75
_models_base = 'https://github.com/GoodAI/goodai-ltm-artifacts/releases/download/models/goodai-ltm-qpm-model'
_pretrained_map = {
    'qpm-distilroberta-01': f'{_models_base}-1145'
}


class AutoTextMatchingModel:
    """
    Factory class for text matching models.
    """

    @staticmethod
    def shared_pretrained(name: str, device: Union[str, torch.device] = None):
        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
        key = f'{name}|{device}'
        model = AutoTextMatchingModel.model_cache.get(key)
        if model is None:
            model = AutoTextMatchingModel.from_pretrained(name, device)
            AutoTextMatchingModel.model_cache[key] = model
        return model

    @staticmethod
    def from_pretrained(name: str, device: Union[torch.device, str] = None) -> BaseTextMatchingModel:
        if device is None:
            device = torch.device('cuda:0') if torch.cuda.is_available() else torch.device('cpu')
        name = name.strip()
        colon_idx = name.find(':')
        if colon_idx == -1:
            url = _pretrained_map.get(name)
            if url is None:
                raise ValueError(f'GoodAI model not found: {name}')
            model_dict = unpickle_downloaded_url(url)
            model: DefaultRerankingCrossEncoder = model_dict['qpmm']
            model.to(device)
            model.zero_grad(set_to_none=True)
            model.eval()
            return model
        model_type = name[:colon_idx]
        model_name = name[colon_idx + 1:]
        if model_type == 'st':
            return SentenceTransformerTextMatchingModel(model_name)
        elif model_type == 'em':
            emb_model = AutoTextEmbeddingModel.shared_pretrained(model_name, device=device)
            return EmbeddingBasedMatchingModel(emb_model, _default_dist_param)
        else:
            raise ValueError(f'Unknown model type: {model_type}')


AutoTextMatchingModel.model_cache = WeakValueDictionary()


import threading

import numpy as np
import torch
from typing import List, Union

from sentence_transformers import SentenceTransformer

from goodai.helpers.torch_helper import param_count
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel

_openai_lock = threading.Lock()


class SentenceTransformerEmbeddingModel(BaseTextEmbeddingModel):
    """
    Text embedding model based on Hugging Face Sentence Transformers.

    https://huggingface.co/sentence-transformers
    """

    def __init__(self, model_name: str, device: Union[torch.device, str] = None, **kwargs):
        self.st = SentenceTransformer(model_name, device=device, **kwargs)
        self.model_name = model_name

    def get_embedding_dim(self) -> int:
        return self.st.get_sentence_embedding_dimension()

    def get_num_retrieval_embeddings(self) -> int:
        return 1

    def get_num_storage_embeddings(self) -> int:
        return 1

    def get_info(self) -> str:
        return f'SentenceTransformer | Name: {self.model_name} | ' \
               f'Params: {param_count(self.st)/1e+6:.4g} million | Dimensions: {self.get_embedding_dim()}'

    def encode(self, sentences: List[str], batch_size: int = 64, show_progress_bar: bool = False,
               convert_to_tensor: bool = False,
               device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        emb = self.st.encode(sentences, batch_size=batch_size, show_progress_bar=show_progress_bar,
                             convert_to_tensor=convert_to_tensor, convert_to_numpy=not convert_to_tensor,
                             device=device, normalize_embeddings=True)
        return emb[:, None, :]

    def encode_queries(self, queries: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                       convert_to_tensor: bool = False,
                       device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode(queries, batch_size=batch_size, show_progress_bar=show_progress_bar,
                           convert_to_tensor=convert_to_tensor, device=device)

    def encode_corpus(self, passages: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                      convert_to_tensor: bool = False,
                      device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode(passages, batch_size=batch_size, show_progress_bar=show_progress_bar,
                           convert_to_tensor=convert_to_tensor, device=device)



import torch
from torch import nn


class EmbCrossProbLossModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.criterion = nn.CrossEntropyLoss()
        self.dist_param = nn.Parameter(torch.ones((1, 1)))

    def forward(self, embeddings_a: torch.Tensor, embeddings_b: torch.Tensor):
        # embeddings_a: (batch_size, num_rk, emb_dim,)
        # embeddings_b: (batch_size, num_sk, emb_dim,)
        batch_size = embeddings_a.size(0)
        num_rk = embeddings_a.size(1)
        num_sk = embeddings_b.size(1)
        embeddings_a = embeddings_a.view(batch_size * num_rk, -1)
        embeddings_b = embeddings_b.view(batch_size * num_sk, -1)
        c_dist = torch.cdist(embeddings_a, embeddings_b)
        # c_dist: (batch_size * num_rk, batch_size * num_sk,)
        c_dist = c_dist.view(batch_size, num_rk, batch_size, num_sk)
        c_dist = torch.amin(c_dist, dim=1)
        # c_dist: (batch_size, batch_size, num_sk,)
        c_dist = torch.amin(c_dist, dim=2)
        # c_dist: (batch_size, batch_size,)
        c_prob_logs = -(c_dist * (5.0 * self.dist_param)).pow(2)
        label = torch.arange(0, batch_size, dtype=torch.long, device=embeddings_a.device)
        return (self.criterion(c_prob_logs, label) + self.criterion(c_prob_logs.transpose(0, 1), label)) / 2

    def get_dist_param_scalar(self) -> float:
        return self.dist_param[0, 0].item()

    @staticmethod
    def get_prob(distances: torch.Tensor, dist_param: float):
        # distances: (batch_size, 1,)
        c_prob_logs = -(distances * dist_param).pow(2)
        return torch.exp(c_prob_logs)



import html
import re


def text_to_html(text: str) -> str:
    h_text = html.escape(text)
    h_text = re.sub(r'\r?\n\r?\n', '<p>', h_text)
    h_text = re.sub(r'\r?\n', '<br>', h_text)
    return h_text



import codecs
import json
import os
from dataclasses import dataclass
from typing import List, Dict, Tuple, Union
import numpy as np
from goodai.helpers.file_helper import download_zip
from goodai.helpers.tokenizer_helper import get_sentence_punctuation_ids
from goodai.ltm.data.query_passage.example import QueryPassageExample
from transformers import PreTrainedTokenizer
from goodai.ltm.data.names import NameSource
from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource


_data_url = 'https://github.com/GoodAI/goodai-ltm-artifacts/releases/download/data/summarized-wikianswers.zip'
_data_file_name = 'summarized-wikianswers.json'


@dataclass
class WikiAnswersEntry:
    questionIds: List[List[int]]
    answerIds: List[int]

    @staticmethod
    def from_example(example: dict, tokenizer: PreTrainedTokenizer) -> 'WikiAnswersEntry':
        questions = example['questions']
        answer = example['answer']
        question_ids = [tokenizer.encode(' ' + q, add_special_tokens=False) for q in questions]
        answer_ids = tokenizer.encode(' ' + answer, add_special_tokens=False)
        return WikiAnswersEntry(question_ids, answer_ids)


class WikiAnswersQueryPassageDataSource(BaseQueryPassageDataSource):
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                 qa_examples: List[dict],
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 min_query_tokens=6):
        super().__init__()
        if max_query_tokens < min_query_tokens:
            raise ValueError(f'Expected max_query_tokens to be at least {min_query_tokens}')
        self.qa_examples = qa_examples
        self.min_query_tokens = min_query_tokens
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.tokenizer = tokenizer
        self.random = random
        self.tokenization_cache: Dict[int, WikiAnswersEntry] = dict()
        self.punctuation_ids = get_sentence_punctuation_ids(tokenizer,
                                                            include_line_break=False, include_colon=True)

    @classmethod
    def create_data_sources(cls, train_fraction: float, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                            max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int) ->\
            Tuple['WikiAnswersQueryPassageDataSource', 'WikiAnswersQueryPassageDataSource']:
        data_dir = download_zip(_data_url)
        data_file = os.path.join(data_dir, _data_file_name)
        with codecs.open(data_file, 'r', 'utf-8') as fd:
            data = json.load(fd)
        random.shuffle(data)
        train_len = round(len(data) * train_fraction)
        train_entries = data[:train_len]
        test_entries = data[train_len:]
        train_ds = cls(random, tokenizer, train_entries, max_query_tokens, min_passage_tokens, max_passage_tokens)
        test_ds = cls(random, tokenizer, test_entries, max_query_tokens, min_passage_tokens, max_passage_tokens)
        return train_ds, test_ds,

    def get_tokenization(self, index: int) -> WikiAnswersEntry:
        tok_entry = self.tokenization_cache.get(index)
        if tok_entry is None:
            index = self.random.randint(0, len(self.qa_examples))
            qa_example: dict = self.qa_examples[index]
            tok_entry = WikiAnswersEntry.from_example(qa_example, self.tokenizer)
            self.tokenization_cache[index] = tok_entry
        return tok_entry

    def get_query_token_ids(self, tok_entry: WikiAnswersEntry, ends_with_name_p=0.75, use_names_p=0.95):
        r = self.random
        name2 = None
        if r.uniform() < use_names_p:
            name1, name2 = tuple(NameSource.get_instance().sample_first_names(r, 2))
        else:
            name1 = r.choice(['User', 'Player'])
        name_lead = name1 + ':'
        name1_ids = self.tokenizer.encode(name_lead, add_special_tokens=False)
        question_ids = tok_entry.questionIds
        nq = len(question_ids)
        if nq == 0:
            return None
        q_idx = r.randint(0, nq)
        selected_q_ids = question_ids[q_idx]
        token_ids = name1_ids + selected_q_ids
        ends_with_name = r.uniform() < ends_with_name_p
        if ends_with_name:
            if name2 is None:
                name2 = r.choice(['AI', 'Assistant', 'Agent', 'NPC'])
            name2_ids = self.tokenizer.encode('\n' + name2 + ':', add_special_tokens=False)
            token_ids += name2_ids
        return token_ids[-self.max_query_tokens:]

    def get_passage_token_ids(self, tok_entry: WikiAnswersEntry):
        r = self.random
        answer_ids = tok_entry.answerIds
        na = len(answer_ids)
        if na < self.min_passage_tokens:
            return None
        excess = na - self.max_passage_tokens
        if excess > 0:
            idx = r.randint(0, excess + 1)
            return answer_ids[idx:idx + self.max_passage_tokens]
        else:
            return answer_ids

    def sample_item(self, pos_index: int, is_match: bool) -> QueryPassageExample:
        r = self.random
        n = len(self.qa_examples)
        for attempt in range(100):
            pos_tok_entry = self.get_tokenization(pos_index)
            query_token_ids = self.get_query_token_ids(pos_tok_entry)
            if query_token_ids is None:
                pos_index = r.randint(0, n)
                continue
            if is_match:
                passage_tok_entry = pos_tok_entry
            else:
                neg_index = None
                for na in range(100):
                    neg_index = r.randint(0, n)
                    if neg_index == pos_index:
                        continue
                    break
                if neg_index is None:
                    raise SystemError('Unable to find suitable example!')
                neg_tok_entry = self.get_tokenization(neg_index)
                passage_tok_entry = neg_tok_entry
            passage_token_ids = self.get_passage_token_ids(passage_tok_entry)
            if passage_token_ids is None:
                if is_match:
                    pos_index = r.randint(0, n)
                continue
            return QueryPassageExample(query_token_ids, passage_token_ids, is_match)
        raise SystemError('Unable to find suitable example!')

    def sample_item_indexes(self, total_items: int, sample_count: int) -> Union[List[int], np.ndarray]:
        replace = sample_count > total_items
        return self.random.choice(range(total_items), size=sample_count, replace=replace)

    def sample_items(self, count: int, approx_positive_fraction: float = 0.5) -> List[QueryPassageExample]:
        rnd_samples = self.random.uniform(size=count)
        pos_indexes = self.sample_item_indexes(len(self.qa_examples), count)
        is_match = rnd_samples <= approx_positive_fraction
        return [self.sample_item(pos_indexes[i], is_match[i]) for i, _ in enumerate(range(count))]



import torch

from goodai.helpers.tokenizer_helper import get_model_inputs
from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource
from torch.utils.data import Dataset
from typing import List, Tuple
from transformers import PreTrainedTokenizer


class QueryPassageDataset(Dataset):
    def __init__(
        self,
        data_sources: List[Tuple[BaseQueryPassageDataSource, float]],
        tokenizer: PreTrainedTokenizer,
        num_examples: int,
        device: torch.device,
        add_special_tokens: bool = True,
        approx_positive_fraction: float = 0.5,
    ):
        super().__init__()
        weight_sum = sum(w for _, w in data_sources)

        self.data_sources = data_sources
        self.tokenizer = tokenizer
        self.num_examples = num_examples
        self.device = device
        pad_id = tokenizer.pad_token_id
        if pad_id is None:
            raise SystemError('Tokenizer has no PAD token.')
        bos_id = tokenizer.bos_token_id
        eos_id = tokenizer.eos_token_id

        label_list = []
        query_list = []
        passage_list = []

        for i, (ds, weight) in enumerate(data_sources):
            w = weight / weight_sum
            n = round(num_examples * w)
            items = ds.sample_items(n, approx_positive_fraction=approx_positive_fraction)
            for item in items:
                query_list.append(item.queryIds)
                passage_list.append(item.passageIds)
                label_list.append([1.0] if item.match else [0.0])

        if add_special_tokens:
            if bos_id is not None:
                query_list = [[bos_id] + seq for seq in query_list]
                passage_list = [[bos_id] + seq for seq in passage_list]
            if eos_id is not None:
                query_list = [seq + [eos_id] for seq in query_list]
                passage_list = [seq + [eos_id] for seq in passage_list]

        query_inputs = get_model_inputs(query_list, pad_id, device, return_token_lengths=True, tokenizer=tokenizer)
        passage_inputs = get_model_inputs(passage_list, pad_id, device)
        self.query_input_ids = query_inputs['input_ids']
        self.query_att_mask = query_inputs['attention_mask']
        self.query_token_lengths = query_inputs['token_lengths']
        self.passage_input_ids = passage_inputs['input_ids']
        self.passage_att_mask = passage_inputs['attention_mask']
        self.labels = torch.as_tensor(label_list, dtype=torch.float, device=device)

    def __len__(self):
        return self.labels.size(0)

    def __getitem__(self, idx):
        q_input_ids = self.query_input_ids[idx]
        q_token_lengths = self.query_token_lengths[idx]
        q_att_mask = self.query_att_mask[idx]
        p_input_ids = self.passage_input_ids[idx]
        p_att_mask = self.passage_att_mask[idx]
        label = self.labels[idx]
        return (q_input_ids, q_token_lengths, q_att_mask), (p_input_ids, p_att_mask), label,



import os
import pickle
from typing import Dict, Any
import boto3
from dotenv import load_dotenv

load_dotenv()


class CloudStorage:
    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name
        aws_key_id = os.environ.get('AWS_ACCESS_KEY_ID')
        if aws_key_id is None:
            raise RuntimeError('AWS_ACCESS_KEY_ID needs to be set')
        aws_secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
        if aws_secret_key is None:
            raise RuntimeError('AWS_SECRET_ACCESS_KEY needs to be set')
        self.client = boto3.client(
            's3',
            aws_access_key_id=aws_key_id,
            aws_secret_access_key=aws_secret_key,
        )

    def put_object(self, key: str, obj: Dict[str, Any]):
        obj_bytes = pickle.dumps(obj)
        self.client.put_object(Bucket=self.bucket_name, Body=obj_bytes, Key=key)

    def put_object_bytes(self, key: str, obj_bytes: bytes):
        self.client.put_object(Bucket=self.bucket_name, Body=obj_bytes, Key=key)

    @classmethod
    def get_instance(cls, bucket_name: str):
        instance_attr = f'__instance_{bucket_name}'
        try:
            return getattr(cls, instance_attr)
        except AttributeError:
            instance = cls(bucket_name)
            setattr(cls, instance_attr, instance)
            return instance
